/* mc8prim.h - primitives for MC8-HSP/DSP-ISA cards */

#include <stdlib.h>
#include <stdio.h>
/* #include <conio.h>*/
#include "mc8prim.h"
#include "../drivers/inout.h"

/* Define the size of the relay buffer for 3:2 conversion */
#define MAXRBUF 12

/* Size of buffer for MemReadL/MemWriteL */
#define LBUFSIZE 60

/*********************************************/
/* IO port utility routines (for remapping)  */
/*********************************************/
void mc8OutW (int lAddr, short wData)
{
	outw(lAddr, (unsigned)wData);
}

short mc8InW(int lAddr)
{
	return ((short)inw(lAddr));
}


/************************************************/
/* mc8Open                                      */
/*   - Allocate and initialze board data struct */
/*   - lAddr - port address for the board       */
/*   - wIrq - irq addr for the board            */
/*   - wForce - set to 1 to always open board,  */
/*              even if communication fails     */
/************************************************/
mc8_struct *mc8Open(long lAddr, int wIrq, int wForce)
{
mc8_struct *pmc8Info;
long lTmp;

pmc8Info = malloc(sizeof(mc8_struct));
if(pmc8Info == NULL) {
	/* Error */
	return(NULL);
}
/* Establish board information in mc8_struct structure */
pmc8Info->lAddr = lAddr;
pmc8Info->wIrq = wIrq;
pmc8Info->wProc = 2;		/* Assume 21062 */
pmc8Info->lMAddrXfer = 0x24fc0;    /* End of PM memory, 21062 */
pmc8Info->wXferMax = 0x40;
pmc8Info->wBoardVer = 110;		/* Default version 1.10 */

/* Initialize COFF pointers to null */
pmc8Info->psCoffHdr = NULL;
pmc8Info->psCoffSym = NULL;
pmc8Info->szCoffStr = NULL;
pmc8Info->psCoffSec = NULL;
pmc8Info->plSec = NULL;

/* Test - is this a 21062 board, a 21061, or a non-DSP board */
/* Write to register IM2 - used for SPORT0, which isn't connected on the board */
mc8WriteDSPReg(pmc8Info, 0x71, 2);

/* Read it back */
mc8ReadDSPReg(pmc8Info, 0x71, &lTmp);

if(lTmp == 2) {
	/* 21062 and 060 have the IM2 register */
	/* Nothing to do, until we need the 060 test */
} else if(lTmp == 1) {
	/* 21061 always returns 1 from IM2 */
	pmc8Info->wProc = 1;		/* Specify 21061 */
	pmc8Info->lMAddrXfer = 0x21fc0;		/* End of PM memory, 21061 */
} else if(lTmp == 0) {
	/* No processor present - will return 0*/
	pmc8Info->wProc = -1;
} else if(lTmp == 0xffffffffL && !wForce) {
	/* Board communications are not working - return NULL */
	free(pmc8Info);
	return(NULL);
}
	
return(pmc8Info);
}


/************************************************/
/* mc8Close                                     */
/*   - Free board data struct and its pointers  */
/************************************************/
int mc8Close(mc8_struct *pmc8Info)
{
if(pmc8Info == NULL) {
	return(RET_ERR);
}

/* Free COFF pointers */
if(pmc8Info->psCoffHdr != NULL) free(pmc8Info->psCoffHdr);
if(pmc8Info->psCoffSym != NULL) free(pmc8Info->psCoffSym);
if(pmc8Info->szCoffStr != NULL) free(pmc8Info->szCoffStr);
if(pmc8Info->psCoffSec != NULL) free(pmc8Info->psCoffSec);
if(pmc8Info->plSec != NULL)     free(pmc8Info->plSec);

/* Free the mc8_struct */
free(pmc8Info);

return(RET_OK);
}


/************************************************/
/* mc8ReadReg                                   */
/*   - read data from MC8-DSP/HSP-ISA register  */
/************************************************/
int mc8ReadReg(mc8_struct *pmc8Info, int wReg, short *pwData)
{

/* Just pass request down - 1 read */
return(mc8ReadRegs(pmc8Info, wReg, pwData, 1));
}


/************************************************/
/* mc8ReadRegs                                  */
/*   - read data from MC8-DSP/HSP-ISA registers */
/************************************************/
int mc8ReadRegs(mc8_struct *pmc8Info, int wReg, short *rgwData, int cRegs)
{
int i;

/* Input qualification */
if(pmc8Info == NULL || rgwData == NULL) {
	/* Error: invalid NULL input */
	return(RET_ERR);
}

if(cRegs <= 0) {
	cRegs = 0;
	return(RET_OK);
}

/* Set address of register (CTRL word = lAddr+2) */
mc8OutW(pmc8Info->lAddr+2, wReg);

/* Read loop */
for(i=0;i<cRegs;i++) {
	rgwData[i] = mc8InW(pmc8Info->lAddr);
}	

return(RET_OK);
}

/************************************************/
/* mc8ReadCtrl                                  */
/*   - read the control word                    */
/************************************************/
mc8ReadCtrl(mc8_struct *pmc8Info, short *pwCtrl)
{
if (pmc8Info == NULL || pwCtrl == NULL) {
	return(RET_ERR);
}

*pwCtrl = mc8InW(pmc8Info->lAddr+2); 

return(RET_OK);
}

/************************************************/
/* mc8WriteReg                                  */
/*   - write data to MC8-DSP/HSP-ISA register   */
/************************************************/
int mc8WriteReg(mc8_struct *pmc8Info, int wReg, short wData)
{

/* Just pass request down - 1 write */
return(mc8WriteRegs(pmc8Info, wReg, &wData, 1));
}


/************************************************/
/* mc8WriteRegs                                 */
/*   - write data to MC8-DSP/HSP-ISA registers  */
/*   - smart write: must fix Host packing bug   */
/************************************************/
int mc8WriteRegs(mc8_struct *pmc8Info, int wReg, short *rgwData, int cRegs)
{
int i,cExtW;
int wAutoIncr,wCur;
int wDSP;


/* Input qualification */
if(pmc8Info == NULL || rgwData == NULL) {
	/* Error: invalid NULL input */
	return(RET_ERR);
}

/* If AutoIncrement bit is set, and access is not to DSP, autoincrement is on */
wAutoIncr = ((wReg & (MC8_AUTOINCR|MC8_DSPREG)) == MC8_AUTOINCR);            

/* Write is to DSP regs - don't check for board writes */
wDSP = ((wReg & MC8_DSPREG) == MC8_DSPREG);

wCur = wReg & 0x7f;		/* Current internal register */

if(cRegs <= 0) {
	cRegs = 0;
	return(RET_OK);
}

/* Set address of register (CTRL word = lAddr+2) */
mc8OutW(pmc8Info->lAddr+2, wReg);

cExtW=0;
/* Write loop */
for(i=0;i<cRegs;i++) {
	mc8OutW(pmc8Info->lAddr, rgwData[i]);
	if(!wDSP) {
		if(wAutoIncr) {
			wCur = (wReg + i) & 0x7f;
		}
		/* Count writes to 'external' ports.  This is *SERIOUSLY*
	   	dependent on the register map, and must change if it does */
		if( (MC8_DAC0 <= wCur && wCur <= MC8_DAC15) ||  		/* DACs */
			(wCur == MC8_DIGOUT0 || wCur == MC8_DIGOUT0H) ||	/* DIGOUT0 */
				/* Decoders: even writes for 0x4n, 0x6n, odd writs 0x5n, 0x7n */
			((wCur >= 0x40) && ((((wCur & 0x70)>>4) + (wCur & 0xf)) % 2) == 1) ) {
		
			cExtW++;
		}
	}
}

if(cExtW > 0 && pmc8Info->wProc != -1 && pmc8Info->wBoardVer < 110) {
	/* If we did external writes, and a DSP is present, clear the bug */
	mc8OutW(pmc8Info->lAddr+2, MC8_DSPREG|DMASTAT);	/* DMASTAT is read only */
	mc8OutW(pmc8Info->lAddr, 0);					/* So writing a 0 can never hurt */
	
	/* If the count is even, need an extra read to get things synched */
	if((cExtW % 2) == 0) {
		mc8InW(pmc8Info->lAddr);
	}
}

return(RET_OK);
}

/************************************************/
/* mc8WriteCtrl                                 */
/*   - write the control word                   */
/************************************************/
mc8WriteCtrl(mc8_struct *pmc8Info, short wCtrl)
{
if(pmc8Info == NULL) {
	/* Error - null input */
	return(RET_ERR);
}
mc8OutW(pmc8Info->lAddr+2,wCtrl); 

return(RET_OK);
}



/*****************************************************/
/* mc8ReadDSPReg                                     */
/*   - read a DSP register                           */
/*****************************************************/
int mc8ReadDSPReg(mc8_struct *pmc8Info, int wReg, long *plData)
{
if(pmc8Info == NULL || plData == NULL) {
	/* Error: Invalid NULL */
	return(RET_ERR);
}

/* Trust that DOS word ordering is correct */
return(mc8ReadRegs(pmc8Info, MC8_DSPREG|wReg, (short *) plData, 2));
}

/*****************************************************/
/* mc8WriteDSPReg                                    */
/*   - write a DSP register                          */
/*****************************************************/
int mc8WriteDSPReg(mc8_struct *pmc8Info, int wReg, long lData)
{
if(pmc8Info == NULL) {
	/* Error: Invalid NULL */
	return(RET_ERR);
}

/* Trust that DOS word ordering is correct */
return(mc8WriteRegs(pmc8Info, MC8_DSPREG|wReg, (short *) (&lData), 2));
}


/*************************************************************/
/* mc8ReadMem                                                */
/*   - read memory from MC8-DSP-ISA                          */
/*   - if addr >= 0x400000, reads external mem via a chunk   */
/*     of on-chip memory (2 DMA accesses)                    */
/*   - cWords is number of SHARC words (x2 or x3 for shorts) */
/*   - Instr vs Data (48/32 bits) done via flag in lAddr     */
/*************************************************************/
int mc8ReadMem(mc8_struct *pmc8Info, long lAddr, short *rgwData, 
			int cWords)
{
int i, i2, wMemSize, wDMAMode, wRet, wIsInst;
long lCount,lMAddr;
short rgsTmp[MAXRBUF];

/* Qualify input */
if(pmc8Info == NULL || rgwData == NULL) {
	/* Error: invalid null */
	return(RET_ERR);
}
/* Qualify cWords */
if(cWords <= 0) cWords = 0;

/* Set memory size - number of short words per SHARC word */
/* This may require interpreting the SYSCON IMDW bits. For now, they're 0 */
if(lAddr & MEMINSTR) {
	wIsInst = 1;
	wMemSize = 3;
	wDMAMode = MC8DMA_IR48;	/* 48 bit reads */
} else {
	wIsInst = 0;
	wMemSize = 2;
	wDMAMode = MC8DMA_IR32; /* 32 bit reads */
}

/* Memory address - without flags */
lMAddr = lAddr & MEMMASK;

/* Check - are we in the internal or external memory */
if(lMAddr >= MC8_EXTADDR) {
	/* External memory must be read to internal, first */
	
	/* If the requested read is too large (for transfer mem block) recurse */
	if(cWords > pmc8Info->wXferMax) {
		for(i=0;i<cWords;i+=pmc8Info->wXferMax) {
			if(i+pmc8Info->wXferMax > cWords) {
				wRet = mc8ReadMem(pmc8Info, lAddr+i, rgwData+i*wMemSize,
					cWords-i);
			} else {
				wRet= mc8ReadMem(pmc8Info, lAddr+i, rgwData+i*wMemSize,
					pmc8Info->wXferMax);
			}
			if(wRet != RET_OK) {
				return(RET_ERR);
			}
		}
		/* After done recursing, exit */
		return(RET_OK);
	}
	
	/* Read from External memory to internal memory */
	/* Set up DMA channel 7 for SRAM->SHARC transfer */
		/* Flush packing state, disable DMA, set new mode */
	mc8WriteDSPReg(pmc8Info, DMAC7, MC8DMA_FLUSH|MC8DMA_EXTR);
		
	/* Load configuration registers */
	/* II7 - Internal Index: 0 => 0x2 0000, etc */
	mc8WriteDSPReg(pmc8Info, II7, (pmc8Info->lMAddrXfer - MC8_INTADDR));
	
	/* IM7 - Modify value */
	mc8WriteDSPReg(pmc8Info, IM7, 1);
	
	/* C7 - Count of internal words */
	mc8WriteDSPReg(pmc8Info, C7, cWords);
	
	/* External configuration */
	/* EI7 - External Index */
	mc8WriteDSPReg(pmc8Info, EI7, lMAddr);
	
	/* EM7 - External Modify */
	mc8WriteDSPReg(pmc8Info, EM7, 1);
	
	/* EC7 - External Count - same as internal (no packing) */
	mc8WriteDSPReg(pmc8Info, EC7, cWords);
	
	/* OK, we're all set to go. */
	/* Initiate transfer - enable DMA */
	mc8WriteDSPReg(pmc8Info, DMAC7, MC8DMA_EXTR|MC8DMA_ENA);
	
	/* Watch for completion of transfer (shouldn't take long!) */
	mc8ReadDSPReg(pmc8Info, EC7, &lCount);
	i=0;
	while(lCount > 0 && i<cWords) {
		mc8ReadDSPReg(pmc8Info, EC7, &lCount);
		i++;	/* Safety timeout */
	}
	
	/* Done!  External memory has been transferred */
	
	/* DMA mode and word width have to be set to wide */
	/* These are only used for the subsequent internal read */
	wMemSize=3;
	wDMAMode=MC8DMA_IR48;

} /* end if(lMAddr >= MC8_EXTADDR) */

/* Read from local memory */
/* Set up DMA channel 7 for SHARC->ISA transfer */
	/* Flush packing state, disable DMA, set new mode */
mc8WriteDSPReg(pmc8Info, DMAC7, MC8DMA_FLUSH|wDMAMode);
		
/* Load configuration registers */
/* II7 - Internal Index: 0 => 0x2 0000, etc */
if(lMAddr >= MC8_EXTADDR) {
	/* Second half of external transfer */
	mc8WriteDSPReg(pmc8Info, II7, (pmc8Info->lMAddrXfer - MC8_INTADDR));
} else {
	/* Normal transfer, Addr 0x20000 maps to 0 */
	mc8WriteDSPReg(pmc8Info, II7, lMAddr-MC8_INTADDR);
}
	
/* IM7 - Modify value */
mc8WriteDSPReg(pmc8Info, IM7, 1);
	
/* C7 - Count of internal words */
mc8WriteDSPReg(pmc8Info, C7, cWords);
	
/* Initiate transfer */
mc8WriteDSPReg(pmc8Info, DMAC7, wDMAMode|MC8DMA_ENA);

/* Read memory from EPB1 */
/* If inital addr was external, we need to do 48-bit transfers */
	/* throw away bottom 16 bits if non-instr data requested */
if(!wIsInst && wMemSize != 2) {
	/* Memory size overridden - data is in 3s */
	i=0;
	/* Read into rgsTmp, and transfer 2s over */
	while(i < cWords*3) {
		if(cWords*3 < i+MAXRBUF) {
			/* Read leftover amount */
			mc8ReadRegs(pmc8Info, MC8_DSPREG|EPB1, rgsTmp, cWords*3-i);
		} else {
			mc8ReadRegs(pmc8Info, MC8_DSPREG|EPB1, rgsTmp, MAXRBUF);
		}
		/* Copy data into rgwData */
		for(i2=0;i2+i<cWords*3 && i2<MAXRBUF;i2++) {
			/* Skip over first word (LSW) */
			if((i2+i)%3==0) continue;
			/* Rely on ints rounding down for the indexing... */
			rgwData[((i2+i)*2)/3] = rgsTmp[i2];
		}
		i+=i2;
	}
	/* Should have filled rgwData appropriately... */
	return(RET_OK);
}

/* Otherwise, reading wMemSize width data, as expected by caller */
return(mc8ReadRegs(pmc8Info, MC8_DSPREG|EPB1, rgwData, cWords*wMemSize));

} /* end mc8ReadMem */


/*************************************************************/
/* mc8WriteMem                                               */
/*   - write memory to MC8-DSP-ISA                           */
/*   - if addr >= 0x400000, writes external mem via a chunk  */
/*     of on-chip memory (2 DMA accesses)                    */
/*   - cWords is number of SHARC words (x2 or x3 for shorts) */
/*   - Instr vs Data (48/32 bits) done via flag in lAddr     */
/*************************************************************/
int mc8WriteMem (mc8_struct *pmc8Info, long lAddr, short *rgwData, 
			int cWords)
{
int i, i2, wMemSize, wDMAMode, wRet, wIsInst;
long lCount,lMAddr;
short rgsTmp[MAXRBUF];

/* Qualify input */
if(pmc8Info == NULL || rgwData == NULL) {
	/* Error: invalid null */
	return(RET_ERR);
}
/* Qualify cWords */
if(cWords <= 0) cWords = 0;

/* Set memory size - number of short words per SHARC word */
/* This may require interpreting the SYSCON IMDW bits. For now, they're 0 */
if(lAddr & MEMINSTR) {
	wIsInst = 1;
	wMemSize = 3;
	wDMAMode = MC8DMA_IW48;	/* 48 bit writes */
} else {
	wIsInst = 0;
	wMemSize = 2;
	wDMAMode = MC8DMA_IW32; /* 32 bit writes */
}

/* Memory address - without flags */
lMAddr = lAddr & MEMMASK;

/* Write to memory, first. */

/* Check - are we in the internal or external memory */
if(lMAddr >= MC8_EXTADDR) {
	/* If the requested write is too large (for transfer mem block) recurse */
	if(cWords > pmc8Info->wXferMax) {
		for(i=0;i<cWords;i+=pmc8Info->wXferMax) {
			if(i+pmc8Info->wXferMax > cWords) {
				wRet = mc8WriteMem(pmc8Info, lAddr+i, rgwData+i*wMemSize,
					cWords-i);
			} else {
				wRet= mc8WriteMem(pmc8Info, lAddr+i, rgwData+i*wMemSize,
					pmc8Info->wXferMax);
			}
			if(wRet != RET_OK) {
				return(RET_ERR);
			}
		}
		/* After done recursing, exit */
		return(RET_OK);
	}
	/* DMA mode and word width have to be set to wide */
	/* These are only used for the subsequent internal read */
	wMemSize=3;
	wDMAMode=MC8DMA_IW48;

} /* end if(lMAddr >= MC8_EXTADDR) */
	

/* Write to internal memory */
/* Set up DMA channel 7 for ISA->SHARC transfer */
	/* Flush packing state, disable DMA, set new mode */
mc8WriteDSPReg(pmc8Info, DMAC7, MC8DMA_FLUSH|wDMAMode);
		
/* Load configuration registers */
/* II7 - Internal Index: 0 => 0x2 0000, etc */
if(lMAddr >= MC8_EXTADDR) {
	/* Relay address */
	mc8WriteDSPReg(pmc8Info, II7, (pmc8Info->lMAddrXfer - MC8_INTADDR));
} else {
	/* Direct write - use specified address (less base addr) */
	mc8WriteDSPReg(pmc8Info, II7, (lMAddr - MC8_INTADDR));
}
	
/* IM7 - Modify value */
mc8WriteDSPReg(pmc8Info, IM7, 1);
	
/* C7 - Count of internal words */
mc8WriteDSPReg(pmc8Info, C7, cWords);
	
/* Initiate transfer */
mc8WriteDSPReg(pmc8Info, DMAC7, wDMAMode|MC8DMA_ENA);

/* Write data to EPB1 */
/* If inital addr was external, we need to do 48-bit transfers */
	/* throw away bottom 16 bits if non-instr data requested */
if(!wIsInst && wMemSize != 2) {
	/* Memory size overridden - have to do 2:3 packing ourselves */
	i=0;
	/* Read into rgsTmp, and transfer 2s over */
	while(i < cWords*3) {
		/* Copy data from rgwData */
		for(i2=0;i2+i<cWords*3 && i2<MAXRBUF;i2++) {
			/* Skip over first word (LSW) */
			if((i2+i)%3==0) {
				rgsTmp[i2]=0;
			} else {
			/* Rely on ints rounding down for the indexing... */
				rgsTmp[i2] = rgwData[((i2+i)*2)/3];
			}
		}
		if(cWords*3 < i+MAXRBUF) {
			/* Write leftover amount */
			mc8WriteRegs(pmc8Info, MC8_DSPREG|EPB1, rgsTmp, cWords*3-i);
		} else {
			mc8WriteRegs(pmc8Info, MC8_DSPREG|EPB1, rgsTmp, MAXRBUF);
		}
		i+=i2;
	}
} else {
	/* Otherwise, writing wMemSize width data, as expected by caller */
	mc8WriteRegs(pmc8Info, MC8_DSPREG|EPB1, rgwData, cWords*wMemSize);
}

if(lMAddr >= MC8_EXTADDR) {
	/* Write to External memory from internal memory */
	/* Set up DMA channel 7 for SHARC->SRAM transfer */
		/* Flush packing state */
	mc8WriteDSPReg(pmc8Info, DMAC7, MC8DMA_FLUSH|MC8DMA_EXTW);
		
	/* Load configuration registers */
	/* II7 - Internal Index: 0 => 0x2 0000, etc */
	mc8WriteDSPReg(pmc8Info, II7, (pmc8Info->lMAddrXfer - MC8_INTADDR));
	
	/* IM7 - Modify value */
	mc8WriteDSPReg(pmc8Info, IM7, 1);
	
	/* C7 - Count of internal words */
	mc8WriteDSPReg(pmc8Info, C7, cWords);
	
	/* External configuration */
	/* EI7 - External Index */
	mc8WriteDSPReg(pmc8Info, EI7, lMAddr);
	
	/* EM7 - External Modify */
	mc8WriteDSPReg(pmc8Info, EM7, 1);
	
	/* EC7 - External Count - same as internal (no packing) */
	mc8WriteDSPReg(pmc8Info, EC7, cWords);
	
	/* OK, we're all set to go. */
	/* Initiate transfer - enable DMA */
	mc8WriteDSPReg(pmc8Info, DMAC7, MC8DMA_EXTW|MC8DMA_ENA);
	
	/* Watch for completion of transfer (shouldn't take long!) */
	mc8ReadDSPReg(pmc8Info, EC7, &lCount);
	i=0;
	while(lCount > 0 && i<cWords) {
		mc8ReadDSPReg(pmc8Info, EC7, &lCount);
		i++;	/* Safety timeout */
	}
	
	/* Done!  External memory has been transferred */
} /* end if(lMAddr >= MC8_EXTADDR) */

return(RET_OK);

} /* end mc8WriteMem */

/********************************************************/
/******** Length-specific memory functions **************/
/********************************************************/

/***************************************************************/
/* mc8ReadMemL                                                 */
/*  - read data from memory into an array of size long         */
/*  - loops, grabbing LBUFSIZE shorts per cycle                */
/***************************************************************/
int mc8ReadMemL(mc8_struct *pmc8Info, long lAddr, short *rgsData, int cWords)
{
long l;
int i,wRet,iData,cWordsCur;
short rgsBuf[LBUFSIZE];

if(lAddr & MEMINSTR) {
	/* Memory target is 'instruction' width - 40/48 bits.  Want data: 32 bits */
	/* Read into a different buffer, then reformat */
	iData = 0;	/* Index into rgsData */
	for(l=0;l<cWords;l+=LBUFSIZE/3) {
		if(cWords-l > (LBUFSIZE/3)) {
			/* Read a bufferful */
			cWordsCur = LBUFSIZE/3;
		} else {
			/* Read the last of the data */
			cWordsCur = (int)((long)cWords-l);
		}
		
		wRet = mc8ReadMem(pmc8Info,lAddr+l,rgsBuf,cWordsCur);
		if(wRet != RET_OK) break;

		for(i=0;i<cWordsCur;i++) {
			rgsData[iData+i*2] = rgsBuf[i*3+1];	/* LSW = middle word */
			rgsData[iData+i*2+1] = rgsBuf[i*3+2];	/* MSW = MSW */
		}
		iData+=cWordsCur*2;
	}
} else { /* end if (lAddr & MEMINSTR) */
	/* Read normally - sizes are compatible */
	wRet = mc8ReadMem(pmc8Info,lAddr,rgsData,cWords);
}

return(wRet);
}


/***************************************************************/
/* mc8WriteMemL                                                */
/*  - write data to memory from an array of size long          */
/*  - loops, sending LBUFSIZE shorts per cycle                 */
/***************************************************************/
int mc8WriteMemL(mc8_struct *pmc8Info, long lAddr, short *rgsData, int cWords)
{
long l;
int i,wRet,iData,cWordsCur;
short rgsBuf[LBUFSIZE];

if(lAddr & MEMINSTR) {
	/* Memory target is 'instruction' width - 40/48 bits.  Want data: 32 bits */
	/* Reformat into a different buffer, then write*/
	iData = 0;	/* Index into rgsData */
	for(l=0;l<cWords;l+=LBUFSIZE/3) {
		if(cWords-l > (LBUFSIZE/3)) {
			/* Write a bufferful */
			cWordsCur = LBUFSIZE/3;
		} else {
			/* Write the last of the data */
			cWordsCur = (int)((long)cWords-l);
		}

		for(i=0;i<cWordsCur;i++) {
			rgsBuf[i*3] = 0;						/* LSW = 0 */
			rgsBuf[i*3+1] = rgsData[iData+i*2];		/* middle word = LSW */
			rgsBuf[i*3+2] = rgsData[iData+i*2+1];	/* MSW = MSW */
		}
		iData+=cWordsCur*2;
		
		wRet = mc8WriteMem(pmc8Info,lAddr+l,rgsBuf,cWordsCur);
		if(wRet != RET_OK) break;

	}
} else { /* end if (lAddr & MEMINSTR) */
	/* Write normally - sizes are compatible */
	wRet = mc8WriteMem(pmc8Info,lAddr,rgsData,cWords);
}

return(wRet);
}



/*******************************************************/
/*           DSP control functions                     */
/*******************************************************/

/**************************************************/
/* mc8DSPInit                                     */
/*   - initializes the DSP to a known state       */
/*   - configures memory bank size and waitstates */
/**************************************************/

int mc8DSPInit(mc8_struct *pmc8Info)
{

/* Check input validity */
if(pmc8Info == NULL) {
	/* Null input */
	return(RET_ERR);
}
if(pmc8Info->wProc == -1) {
	/* No CPU */
	return(RET_ERR);
}

/* Clear RESET bit */
mc8WriteReg(pmc8Info, MC8_DSPRESET, 0);

/* Write SYSCON = 0x6810 - 512k word memory banks, BHD=1, 16:32 packing */
mc8WriteDSPReg(pmc8Info, SYSCON, 0x6810L);

/* Write WAIT = 21ad:69c1: Ext bank 0 (mem): 0 waits, internal only, */
/*                         Ext bank 1 (board): 3 waits, both time&ACK */
mc8WriteDSPReg(pmc8Info, WAIT, 0x21ad69c1L);

return(RET_OK);
}


/**************************************************/
/* mc8DSPReset                                    */
/*   - soft reset of the on-board processor       */
/*   - will work even if host packing out of sync */
/*   - reconfigures the processor once done       */
/**************************************************/
int mc8DSPReset(mc8_struct *pmc8Info)
{
short rgwData[3], i;

/* Check input validity */
if(pmc8Info == NULL) {
	/* Null input */
	return(RET_ERR);
}
if(pmc8Info->wProc == -1) {
	/* No CPU */
	return(RET_ERR);
}

/* Clear RESET bit */
mc8WriteReg(pmc8Info, MC8_DSPRESET, 0);

/* Write SYSCON = 1 3 times - guarantees reset, even if */
/* the host packing mode is in a weird state */
rgwData[0] = 1;
rgwData[1] = 1;
rgwData[2] = 1;
mc8WriteRegs(pmc8Info, MC8_DSPREG|SYSCON, rgwData, 3);

/* Read SYSCON, looking for 0x10 (then read once more) */
for(i=0;i<6;i++) { /* only read at most 6 times */
	mc8ReadReg(pmc8Info, MC8_DSPREG|SYSCON, rgwData);
	if(rgwData[0] == 0x10) 
		break;
}

if(i<6) {
	/* Get host packing into right synch */
	mc8ReadReg(pmc8Info, MC8_DSPREG|SYSCON, rgwData);
} else {
	/* If we didn't see 0x10, the system is screwed up */
	return(RET_ERR);
}

/* Reconfigure the processor */
mc8DSPInit(pmc8Info);

return(RET_OK);
}


/**************************************************/
/* mc8DSPHardReset                                */
/*   - hardware reset of the on-board processor   */
/*     WARNING: Will reset decoders in v1.0 board */
/*   - reconfigures the processor once done       */
/**************************************************/

int mc8DSPHardReset(mc8_struct *pmc8Info)
{

/* Check input validity */
if(pmc8Info == NULL) {
	/* Null input */
	return(RET_ERR);
}
if(pmc8Info->wProc == -1) {
	/* No CPU */
	return(RET_ERR);
}

/* Set RESET bit - resets the DSP and the quadrature decoders */
mc8WriteReg(pmc8Info, MC8_DSPRESET, 1);

/* Clear the RESET bit again */
mc8WriteReg(pmc8Info, MC8_DSPRESET, 0);

/* DSP is now reset */

/* Reconfigure the processor */
mc8DSPInit(pmc8Info);

return(RET_OK);
}


/**************************************************/
/* mc8DSPStart                                    */
/*   - kicks DSP out of post-reset hold           */
/*     by reading from DMA channel 6              */
/*   - reads and writes 0x20005 and 0x20040 to    */
/*     ensure the DSP starts right (rti at 20040) */
/**************************************************/
int mc8DSPStart(mc8_struct *pmc8Info)
{
long lData;
short rgsData[12];

/* Check input validity */
if(pmc8Info == NULL) {
	/* Null input */
	return(RET_ERR);
}
if(pmc8Info->wProc == -1) {
	/* No CPU */
	return(RET_ERR);
}

/* Clear RESET bit - make sure SHARC can hear us */
mc8WriteReg(pmc8Info, MC8_DSPRESET, 0);

/* rgsData[0-2] = jump (pc,0): 0x73e0 0000 0000 */
rgsData[0]=0;
rgsData[1]=0;
rgsData[2]=0x073e;
/* rgsData[3-5] = rti: 0x0b3e 0000 0000 (LSW first) */
rgsData[3]=0;
rgsData[4]=0;
rgsData[5]=0x0b3e;

/* Read from 0x20005 into rgsData[6-8] (instr width) */
mc8ReadMem(pmc8Info, 0x20005L|MEMINSTR, rgsData+6, 1);
/* Replace 0x20005 with a jump(pc,0) */
mc8WriteMem(pmc8Info, 0x20005L|MEMINSTR, rgsData+0, 1);

/* Read from 0x20040 into rgsData[9-11] (instr width) */
mc8ReadMem(pmc8Info, 0x20040L|MEMINSTR, rgsData+9, 1);
/* Replace 0x20040 (EPB0 isr) with rti */
mc8WriteMem(pmc8Info, 0x20040L|MEMINSTR, rgsData+3, 1);

/*** Jump the processor out of post-reset hold ***/
/* Write 32-bit read command to DMAC6 (disabled) */
mc8WriteDSPReg(pmc8Info, DMAC6, MC8DMA_IR32);

/* Write '1' to C6 - read one value */
mc8WriteDSPReg(pmc8Info, C6, 1);

/* Write read command and enable to DMAC6 */
mc8WriteDSPReg(pmc8Info, DMAC6, MC8DMA_IR32|MC8DMA_ENA);

/* Read the value back, just to keep things stable */
mc8ReadDSPReg(pmc8Info, EPB0, &lData);

/* And the DSP should be up and running, looping at 0x20005 */

/* Fix 0x20040 - write from rgsData[9-11] (instr width) */
mc8WriteMem(pmc8Info, 0x20040L|MEMINSTR, rgsData+9, 1);

/* Fix 0x20005 - write from rgsData[6-8] (instr width) */
mc8WriteMem(pmc8Info, 0x20005L|MEMINSTR, rgsData+6, 1);

/* Now, the SHARC should be off and running */

/* Write twice more to flush the ADSP2106x Shadow Write FIFO */
/* Otherwise, a reset could restore the previous values in mem. */
mc8WriteMem(pmc8Info, 0x20005L|MEMINSTR, rgsData+6, 1);
mc8WriteMem(pmc8Info, 0x20005L|MEMINSTR, rgsData+6, 1);

return(RET_OK);
}
