/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*=~=~=~=~ McGill-Computer-Vision-and-Robotics-Laboratory--C-Module ~=~=~=~=
 *		
 * MODULE NAME:	singularity.c
 *
 * DESCRIPTION:	Manipulator dependant functions that are used in testing
 *		kinematic code.
 *
 *		A NOTE, BEFORE THE DESCRIPTION, ABOUT UNITS:
 *
 *		   All rotational units are radians; all translational units
 *		   are millimters.
 *
 *		elbowSingularityDetect (angles, distance, kyn) is a function 
 *	 	   that 
 *		   returns a bit mask describing any singularities that the 
 *		   manipulator is at or near. The arm is defined to be near
 *		   a singularity when associated angle is at or within 
 *		   'distance' units of the singular position. 'Angles' is 
 *		   an array of 6 floats.
 *
 *		elbowSingularitySet (outangles, inangles, distance, mask, 
 *    			            kyn, jls) 
 *		   copies 'inangles' to 'outangles', except for making
 *		   the minimum changes necessary to ensure that the 
 *		   manipulator is 'distance' units away from the singularities
 *		   specified by the mask. If there is more than one choice
 *		   of angle to use to achieve the singularity, the one
 *		   closest to the original angle is used.
 *
 *		   This routine can be used to put the manipulator at a
 *		   singularity ('distance = 0'), or some arbitrary distance
 *		   away from it. 
 *
 *		   WARNING: The routine is not too smart: If 'distance' is
 *		   large enough to cause the manipulator to be at ANOTHER
 *		   sinularity, this will not be detected.
 * 
 * AUTHOR/DATE:	John Lloyd, Spring 87
 *
 *~=~=~=~=~=~= McGill-Computer-Vision-and-Robotics-Laboratory ~=~=~=~=~=~=*/

#include <math.h>

#include "../h/cdefs.h"
#include "../h/robot_jls.h"
#include "../h/robotMath.h"
#include "../h/elbow_kynvar.h"

/* ---- defines for 'kyn' constants, to make code easier to read ---- */

#define A2		ekyn->gen->dh[1].a	/* basic A matrix parameters */
#define A3		ekyn->gen->dh[2].a
#define A4		ekyn->gen->dh[3].a

static float range180(a)
float a;
{
	/* puts a in the range (-PI,PI] */

	while (a <= -PI)
	 { a += PIT2;
	 }
	while (a > PI)
	 { a -= PIT2;
	 }
	return ( (a > -PI) ? a : PI );
}

static float range360(a)
float a;
{
	/* puts a in the range [0,2 PI) */

	while (a < 0.0)
	 { a += PIT2;
	 }
	while (a >= PIT2)
	 { a -= PIT2;
	 }
	return ( (a >= 0.0) ? a : 0.0 );
}

elbowSingularityDetect (angles, distance, kyn)
float *angles;
float distance;
KYN *kyn;
{
	ELBOW_KYN *ekyn = (ELBOW_KYN*)kyn->ext;

	int code = 0;
	float phi;
	float si3, co3, si34, co34;

	distance = FABS(distance);

	if (closeTo (angles[4], 0.0, distance) ||
	    closeTo (angles[4], PI, distance))
	 { code |= ELBOW_WRIST_DEG;
	 }
	if (closeTo (angles[2], 0.0, distance) ||
	    closeTo (angles[2], PI, distance))
	 { code |= ELBOW_ELBOW_DEG;
	 }

	si3 = sin (angles[2]);
	co3 = cos (angles[2]);
	si34 = sin (angles[2]+angles[3]);
	co34 = cos (angles[2]+angles[3]);

	phi = atan2 (A4*si34 + A3*si3, A4*co34 + A3*co3 + A2);

	if (closeTo (angles[1]+phi,  PI/2, distance) || 
	    closeTo (angles[1]+phi, -PI/2, distance))
	 { code |= ELBOW_ALIGN_DEG;
	 }

	return (code);
}

elbowSingularitySet (outangles, inangles, distance, mask, kyn, jls)
float outangles[];
float inangles[];
float distance;
unsigned long mask;
KYN *kyn;
JLS *jls;
{
	ELBOW_KYN *ekyn = (ELBOW_KYN*)kyn->ext;

	int j;
	float si3, co3, si34, co34, phi;
	float angA, angB;

	for (j=0; j<ELBOWJ; j++)
	 { outangles[j] = inangles[j];
	 }

	if (mask & ELBOW_WRIST_DEG)
	 { outangles[4] = distance;
	 }

	if (mask & ELBOW_ELBOW_DEG)
	 { outangles[2] = distance;
	 }

	if (mask & ELBOW_ALIGN_DEG)
	 { si3 = sin (outangles[2]);
	   co3 = cos (outangles[2]);
	   si34 = sin (outangles[2]+outangles[3]);
	   co34 = cos (outangles[2]+outangles[3]);

	   phi = atan2 (A4*si34 + A3*si3, A4*co34 + A3*co3 + A2);

	   /* Map the angles into range coordinates so we can compare them */
	   
	   angA = PI/2 - phi + distance;
	   angB = -PI/2 - phi + distance;
	   if (angA < jls->jmin[1] || angA > jls->jmax[1])
	    { outangles[1] = angB;
	    }
	   else
	    { outangles[1] = angA;
	    }
	 }
}

closeTo (angle, ref, delta)
float angle, ref, delta;
{
	/* Returns true if 'angle' is within 'ref' of 'delta' */

	float theta;

	theta = range180 (angle - ref);
	if (FABS(theta) <= FABS(delta))
	 { return (1);
	 }
	else
	 { return (0);
	 }
}

#ifdef TEST

main()
{
	float angles[ELBOWJ];
	float result[ELBOWJ];
	float output[ELBOWJ];
	int mask, i;

	extern char* getDefaultRobot();
	KYN *kyn;
	JLS *jls;

	if ((jls = getJls(getDefaultRobot())) == NULL)
	 { printErrors();
	   exit (-1);
	 }
	if ((kyn = getKyn(getDefaultRobot(), NULL)) == NULL)
	 { printErrors();
	   exit (-1);
	 }

	while (1)
	 { 
	   printf ("Enter angles and mask> ");
	   scanVf (angles, 6);
	   scaleVf (angles, DEGTORAD, angles, 6);
	   scanf ("%x", &mask);
	   while (getchar() != '\n')
	      ;

	   elbowSingularitySet (result, angles, DEGTORAD*.2, mask, kyn, jls);

	   scaleVf (output, RADTODEG, result, 6);
	   printVf ("%10.4f\n", output, 6);

	   printf ("Detection with espsilon .2: 0x%x\n",
		elbowSingularityDetect (result, DEGTORAD*(.2+.0001), kyn));
	 }
}

#endif
