/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd and Vincent Hayward.	    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*=~=~=~=~ McGill-Computer-Vision-and-Robotics-Laboratory--C-Module ~=~=~=~=
 *		
 * MODULE NAME:	gantryKinematic.c
 *
 * DESCRIPTION:	Forward and inverse kinematic routines for the GANTRY robots.
 * 
 * AUTHOR/DATE:	John Lloyd, Spring 1987, after Vincent Hayward, Purdue, 1983
 *		Rewrote routines to interface with new RCI, winter 1988.
 *
 *~=~=~=~=~=~= McGill-Computer-Vision-and-Robotics-Laboratory ~=~=~=~=~=~=*/

#include <math.h>

#include "../h/fastmath.h"
#include "../h/cdefs.h"
#include "../h/robot_jls.h"
#include "../h/robotMath.h"
#include "../h/gantry_jls.h"
#include "../h/gantry_kynvar.h"
#include "../h/robotTypes.h"

#include "../h/rcclDefs.h"	/* for F_UNDEF */

#define EPSILONF 2.0e-6

#define SGN(x)  ((x) >= 0 ? 1 : -1)

/* performs mod 2 pi */

static float range(a) /*##*/	/* Sets a to  0 <= a < pi2  */
double a;
{
	if (a > 15. || a < -15.)
	 { return(10.);
	 }

	while(a >= PIT2)
	 { a -= PIT2;
	 }
	while(a < 0.)
	 { a += PIT2;
	 }
	if (a > PIT2)
	 { return(0.);
	 }
	return(a);
}

/* ---- Definition of link parameters to make code more readable ---- */

#define J1	0
#define J2	1
#define J3	2
#define J4	3
#define J5	4
#define J6	5

#define S2	skyn->s2
#define C2	skyn->c2
#define SA	skyn->sa
#define CA	skyn->ca

gantryFwdKinematics (t, conf, j, kyn, var)
TRSF *t;	  	  /* resulting transform */
unsigned long *conf;	  /* returns configuration information */
float *j;		  /* joints in range coordinates + configuration */
KYN *kyn;	  	  /* kinematic parameters */
void *var;		  /* palce for storing computing sines/cosines */
{
	register GANTRY_KYN *skyn = (GANTRY_KYN*)kyn->ext;
	register GANTRY_VAR *svar = (GANTRY_VAR*)var;
	GEN_KYN *gen = kyn->gen;
	TRSF R;
	VECT p1, p2;
	double a4c4, a4s4;

	/* sines and cosines */

	float theta4, c4, s4, c5, s5;

	theta4 = j[J4] + gen->dh[J4].theta;

	SINCOS (theta4, &s4, &c4);
	SINCOS (j[J5], &s5, &c5);

	identTrsf (&R);
	multTrsfRotZ (&R, RADTODEG*theta4);
	multTrsfRotX (&R, RADTODEG*gen->dh[J4].alpha);
	multTrsfRotZ (&R, RADTODEG*j[J5]);
	multTrsfRotX (&R, RADTODEG*gen->dh[J5].alpha);

	multTrsf (&R, &skyn->R03, &R);

	t->n = R.n;
	t->o = R.o;
	t->a = R.a;

	combineVect (&p1, j[J1], &skyn->axis1, j[J2], &skyn->axis2);
	combineVect (&p1, j[J3], &skyn->axis3, 1, &p1);
	
	a4c4 = gen->dh[J4].a*c4;
	a4s4 = gen->dh[J4].a*s4;
	combineVect (&p2, a4c4, &skyn->R03.n, a4s4, &skyn->R03.o);

	addVect (&t->p, &p1, &p2);

	if (conf != NULL)
	 { *conf = 0;
	 }

	if (var != NULL)	/* Update sines and cosinesd if needed */
	 { svar->s4 = s4;
	   svar->c4 = c4;
	   svar->s5 = s5;
	   svar->c5 = c5;
	   svar->d[J1] = j[J1];
	   svar->d[J2] = j[J2];
	   svar->d[J3] = j[J3];
	 }
}

static double epsilon5 = 5.0*DEGTORAD;

gantryInvKinematics(j, t, conf, refj, kyn, jls, var) /*::*/
float *j;	  	  /* joints in range coordinates + configuration */
TRSF *t;	  /* input transform 			  	 */
unsigned long conf;	  /* configuration information			 */
float *refj;		  /* reference joints				 */
KYN *kyn;		  /* kinematic parameters			 */
void *var;		  /* run-time variables 			 */
JLS *jls;		  /* joint level data -- usd to get min. joints	 */
{
	GANTRY_KYN *skyn = (GANTRY_KYN*)kyn->ext;
	register GANTRY_VAR *svar = (GANTRY_VAR*)var;
	GEN_KYN *gen = kyn->gen;

	float ang[GANTRYJ];			/* non-ranged angle values */
	int code = 0;
	int wholeMethod = 1;

	double alpha4, m, phi, theta4, s4, c4, a4;
	VECT u, p2;
	TRSF R, Ru;

	R = *t;
	R.p.x = R.p.y = R.p.z = 0;
	multLiTrsf (&R, &skyn->R03, &R);
	Ru.p.x = Ru.p.y = Ru.p.z = 0;

	alpha4 = gen->dh[J4].alpha;
	m = SGN(alpha4)*sqrt(R.o.x*R.o.x + R.o.y*R.o.y);
	phi = atan2(m, R.o.z);
	if (ABS(m) < EPSILONF)
	 { /* degenerate; use existing theta_4 */
	   if (refj != NULL)
	    { theta4 = refj[J4] + gen->dh[J4].theta;
	    }
	   else
	    { /* no refj; use a value close to mid range value */
	      theta4 = -PI/2;
	    }
	   u.x = -cos(theta4)*SGN(alpha4);
	   u.y = -sin(theta4)*SGN(alpha4);
	   u.z = 0;
	 }
	else
	 { u.x = -R.o.y/m;
	   u.y =  R.o.x/m;
	   u.z = 0;
	 }
	rotToTrsf (&Ru, &u, alpha4-phi);
	multTrsf (&R, &Ru, &R);

	theta4 = atan2 (SGN(alpha4)*R.o.x, -SGN(alpha4)*R.o.y);
	ang[J4] = theta4 - gen->dh[J4].theta;
	ang[J5] = atan2 (SGN(alpha4)*R.n.z, -SGN(alpha4)*R.a.z);

	/* Now compute the translation */

	s4 = sin(theta4);
	c4 = cos(theta4);
	a4 = gen->dh[J4].a;
	combineVect (&p2, a4*c4, &skyn->R03.n, a4*s4, &skyn->R03.o);
	subtractVect (&p2, &t->p, &p2);

	ang[J3] = (p2.z)/skyn->axis3.z;
	ang[J2] = (p2.y - ang[J3]*skyn->axis3.y)/skyn->axis2.y;
	ang[J1] = (p2.x - ang[J2]*skyn->axis2.x - ang[J3]*skyn->axis3.x);

	code |= gantryRangeAngles (j, ang, refj, kyn, jls);

	if (var != NULL)	/* Update sines and cosinesd if needed */
	 { svar->s4 = s4;
	   svar->c4 = c4;
	   svar->s5 = sin(j[J5]);
	   svar->c5 = cos(j[J5]);
	   svar->d[J1] = j[J1];
	   svar->d[J2] = j[J2];
	   svar->d[J3] = j[J3];
	 }
	return(code);
}

gantryRangeAngles (r, a, refj, kyn, jls)
register float *r, *a;
float *refj;
KYN *kyn;
JLS *jls;
{
	register float *jmin, *jmax;
	register double angle;
	register double refAngle;
	int code = 0;

	jmin = jls->jmin;
	jmax = jls->jmax;

	if (a[J1] <= jmin[J1] || a[J1] >= jmax[J1])
	 { code |= (1 << 0);
	 }
	r[J1] = a[J1];

	if (a[J2] <= jmin[J2] || a[J2] >= jmax[J2])
	 { code |= (1 << 1);
	 }
	r[J2] = a[J2];

	if (a[J3] <= jmin[J3] || a[J3] >= jmax[J3])
	 { code |= (1 << 2);
	 }
	r[J3] = a[J3];

	/* Range of joint 4 < 360 degrees */

	if ((angle = a[J4]) <= jmin[J4])
	 { do
	    { angle += 2*PI;
	    }
	   while (angle <= jmin[J4]);
	   if (angle >= jmax[J4])
	    { code |= (1 << 3);
	    }
	 }
	else if (angle >= jmax[J4])
	 { do
	    { angle -= 2*PI;
	    }
	   while (angle >= jmax[J4]);
	   if (angle <= jmin[J4])
	    { code |= (1 << 3);
	    }
	 }
	r[J4] = angle;

	/* Range of joint 4 < 360 degrees */

	if ((angle = a[J5]) <= jmin[J5])
	 { do
	    { angle += 2*PI;
	    }
	   while (angle <= jmin[J5]);
	   if (angle >= jmax[J5])
	    { code |= (1 << 4);
	    }
	 }
	else if (angle >= jmax[J5])
	 { do
	    { angle -= 2*PI;
	    }
	   while (angle >= jmax[J5]);
	   if (angle <= jmin[J5])
	    { code |= (1 << 4);
	    }
	 }
	r[J5] = angle;
	return (code);
}

gantrySolveConf(conf, j, kyn)
unsigned long *conf;	/* configuration description */
register float *j;	/* joints in radians */
KYN *kyn;
{
	GEN_KYN *gen = kyn->gen;

	*conf = 0;
/*  	if (range(j[J5]) > PI) */
/*  	 { *conf |= GANTRY_FLIP_CONF; */
/*  	 } */
}
