/*
  Source Code for ATI ISA FT adapter
 
  Adapted for RCCL from source provided by ATI, written by Jerry Dahl
*/

#include "../drivers/inout.h"

#define  FALSE		0
#define  TRUE		1

static unsigned addr_reg	= 0x280;      /* base I/O address of FT Card */
static unsigned data_reg	= 0x282;      /* addr_reg + 2 */
static unsigned status_reg	= 0x284;      /* addr_reg + 4 */
static unsigned data_available	= FALSE;      /* flag from interrupt handler */

/*------------------------------------------------------------------
   (C) Copyright 1997,  by ATI Industrial Automation
  
   function:  send_command
  
   Uses global registers:
   addr_reg	PC I/O address of F/T card's address register
  	data_reg	PC I/O address of F/T card's data register
  
   Calling parameters:
  	unsigned *bptr 	= pointer to command structure
  	int numwords	= length of command
  
   Return Value: 	1 	if successful
  						0 	if abort
  
  ------------------------------------------------------------------ */
static int send_command(unsigned *bptr, int numwords)
{
	unsigned dword, command, retcode, loop, spin;

	/*---------------------------------------
	  write command block to DPRAM at offset 0x000
	 ---------------------------------------*/
	command = *bptr;	     /* save command word */
	outw(addr_reg,0); 	     /* set address register to command_block*/
	for(;numwords;numwords--)    /* write command block to dual port RAM*/
		{
		dword = *bptr ;
		outw(data_reg,dword) ;
		bptr ++ ;
		}
	/*---------------------------------------
	   read MailBox at DPRAM 0x3FE to clear host interrupt flag
	  ---------------------------------------*/
	outw(addr_reg,0x3FE);	     	/* set mailbox address */
	inw(data_reg);		   	/* reset interrupt flag */

	outw(addr_reg,0x3FE) ;	    	/* set mailbox address */
	outw(data_reg,command);	    	/* clear acknowledgment */

	/*---------------------------------------
	   write command word to Mailbox at DPRAM 0x3FF to set
	   DSP interrupt flag
	  ---------------------------------------*/
	outw(addr_reg,0x3ff) ;
	outw(data_reg,command) ;

	/*---------------------------------------
	    waits for command completion (interrupt flag) from DSP
	--------------------------------------- */
	retcode = TRUE ;
	for (spin=10000; spin; spin--)
		{
		dword = inw(status_reg);

		/* wait for interrupt flag */
		if((dword & 0x0010)==0)       	/* interrupt set */
			{

			/* read mailbox location at 0x3FE to reset 
			   interrupt flag */
			outw(addr_reg,0x3FE) ;	 /* write dpram address*/
			retcode = inw(data_reg); /* reset interrupt flag */

			/* test mailbox for command acknowledgment */
			if(retcode & 0x00ff)
			   { break;
			   }
			}
		}
	if (spin == 0)
	 { return -1;
	 }
	else
	 { return (retcode);
	 }
}

/*------------------------------------------------------------------
   (C) Copyright 1997,  by ATI Industrial Automation
   function:  receive_command
   reads numwords words from ft card starting at fifo offset 0
  
   global registers:
      addr_reg	PC I/O address of F/T card's address register
  	data_reg	PC I/O address of F/T card's data register
  
  
    Calling parameters:
  	unsigned *bptr 	= pointer to command structure
  	int numwords	= length of command
  
  ------------------------------------------------------------------*/
static void receive_command(unsigned *bptr, int numword)
{
	unsigned  dword;

	outw(addr_reg,0);	  /* reset address */
	for(;numword;numword--)
		{
		dword = inw(data_reg);
		*bptr = dword ;
		bptr ++ ;
		}
}


/*-------------------------------------------------
   sensor_data command structure
  -------------------------------------------------*/
struct
	{
	unsigned  	cmd_rc;			/* 0x0100 */
	unsigned  	mode;
	unsigned  	matrix[96];  		/* 6 x 8 32-bit words */
	unsigned		sbias[8];	/* hardware bias array */
	signed  		Mm[8];
	signed  		Bm[8];
	signed  		Mb[8];
	signed  		Bb[8];
	unsigned  	number_of_gages;
	unsigned  	counts_per_force;
	unsigned  	counts_per_torque;
	unsigned  	max_force;
	unsigned  	max_torque;
	unsigned  	maxtemperature;
	unsigned  	min_temperature;
	unsigned  	units;
	unsigned  	serial_number;
	} sensor_data;

/*------------------------------------------------------------------
   (C) Copyright 1997,  by ATI Industrial Automation
  
   function: read_sensor_info()
  	Sends Sensor_Command to F/T card
  	Returns sensor_data
  
   global registers:
      addr_reg	PC I/O address of F/T card's address register
  	data_reg	PC I/O address of F/T card's data register
    Return Value: 	1 	if successful
  				0 	if abort
  
  -----------------------------------------------------------------*/
int atiSensorStart(void)
{
	sensor_data.cmd_rc = 0x0100 ;		/* sendor data command word */
	sensor_data.mode   = 0x0000 ;		/* mode=read */
	if(send_command(&sensor_data,2) <= 0) return(-1);
	receive_command(&sensor_data,sizeof(sensor_data));
	return (TRUE);
}

/*------------------------------------------------------------------
   (C) Copyright 1997,  by ATI Industrial Automation
  
   read_ft_data()
  
   global registers:
      addr_reg	PC I/O address of F/T card's address register
  	data_reg	PC I/O address of F/T card's data register
  
   polls data_reg[0] for new data
  ------------------------------------------------------------------*/

int atiSensorRead(float *data, int *moncode)
{
	int  	loop, i;
	int  	dword ;
	int	force_counts[3], torque_counts[3];

	/*-----------------------------------
	   polling mode, test for new data
	  -----------------------------------*/
	outw(addr_reg,0x100);	/* write ram starting addr */
	dword = inw(data_reg);	/* read first word[0] */
	if(!dword)
	 { return (0);
	 }

	/*-----------------------------------------
	   new data available.
	   first check monitoring conditions
	  -----------------------------------------*/
	dword = inw(data_reg);      /* read 2nd word=monitor cond */
	if(dword)
	 {
	   outw(addr_reg,0x101);  /* monitor condition */
	   outw(data_reg,0x00);	  /* clear monitor condition */
	   /*-----------------------------------
	      mon_cond_status
	      1 = Alarm state
	      2 = Out of Alarm State
	      3 = Expired
	     ------------------------------------*/
	 }

	/*-----------------------------------------
	   get the new F/T data in counts
	   to get actual forces and torque, the application
	   should divide the counts by the
	   counts_per_force and counts_per_torque
	   conversion factors
	  -----------------------------------------*/
	for(i=0;i<3;i++)
	  { force_counts[i] = inw(data_reg);
	    data[i] = 10*(signed short)force_counts[i]/
		      (double)sensor_data.counts_per_force;
	  }
	for(i=0;i<3;i++)
	  { torque_counts[i] = inw(data_reg);
	    data[i+3] = 10*(signed short)torque_counts[i]/
	    		(double)sensor_data.counts_per_torque;
	  }

	/* acknowledge data */
	outw(addr_reg,0x100);   /* write ram starting addr */
	outw(data_reg,0x00);	/* acknowledge data */
	if (moncode) *moncode = dword;
	return (1);
}
