/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*---------------------------------------------------------------------------
 *		
 * MODULE NAME:	jointBlend
 *
 * DESCRIPTION:	This file contains code for doing the blending of joint
 *		angles trajectories.
 *
 *      	The blending is done simply by using Hermite polynominals.
 *      	The function applyJointBlend() adds a correction dJ 
 *      	to an input setpoint j to smooth away discontinuities
 *      	in velocity and acceleration:
 *
 *      		j(output) = j(input) + dJ
 *
 *      	dJ is computed from a 5th degree polynominal:
 *
 *      					2       3       4       5
 *      		dJ(t) = a0 + a1 t + a2 t  + a3 t  + a4 t  + a5 t
 *
 *      	(This is the same polynominal used for blending the 
 *      	 translation vector of Cartesian motions).
 *
 *      	The polynominal coefficients are computed in stages
 *      	by the functions setJointBlend1() and setJointBlend2(),
 *      	which take as inputs sample points from the two paths to
 *      	be blended together.
 *
 *      	The starting acceleration is assumed to be zero ... not quite 
 *      	right, but it works OK.
 *
 * AUTHOR/DATE:	John Lloyd, October 1988
 *
 * MODIFICATION HISTORY:
 *
 *--------------------------------------------------------------------------*/

#include "../h/rcclDefs.h"
#include "jointBlend.h"

void setJointBlend1 (JOINT_BLEND_PARAMS *jbp, JNTS *j1, JNTS *j2, int numJ)
{
	int j;

	/* Compute second part of Hermite correction polynomial */

	for (j=0; j<numJ; j++)
	 { jbp->A0.v[j] = j1->v[j] - j2->v[j];
	   jbp->A3.v[j] = -10.0 * jbp->A0.v[j];
	   jbp->A4.v[j] =  15.0 * jbp->A0.v[j];
	   jbp->A5.v[j] =  -6.0 * jbp->A0.v[j];
	 }
#if 0
	jbp->A0.v[1] = j1->v[1] - j2->v[1];
	jbp->A0.v[2] = j1->v[2] - j2->v[2];
	jbp->A0.v[3] = j1->v[3] - j2->v[3];
	jbp->A0.v[4] = j1->v[4] - j2->v[4];
	jbp->A0.v[5] = j1->v[5] - j2->v[5];

	jbp->A3.v[0] = -10.0 * jbp->A0.v[0];
	jbp->A3.v[1] = -10.0 * jbp->A0.v[1];
	jbp->A3.v[2] = -10.0 * jbp->A0.v[2];
	jbp->A3.v[3] = -10.0 * jbp->A0.v[3];
	jbp->A3.v[4] = -10.0 * jbp->A0.v[4];
	jbp->A3.v[5] = -10.0 * jbp->A0.v[5];

	jbp->A4.v[0] =  15.0 * jbp->A0.v[0];
	jbp->A4.v[1] =  15.0 * jbp->A0.v[1];
	jbp->A4.v[2] =  15.0 * jbp->A0.v[2];
	jbp->A4.v[3] =  15.0 * jbp->A0.v[3];
	jbp->A4.v[4] =  15.0 * jbp->A0.v[4];
	jbp->A4.v[5] =  15.0 * jbp->A0.v[5];

	jbp->A5.v[0] =  -6.0 * jbp->A0.v[0];
	jbp->A5.v[1] =  -6.0 * jbp->A0.v[1];
	jbp->A5.v[2] =  -6.0 * jbp->A0.v[2];
	jbp->A5.v[3] =  -6.0 * jbp->A0.v[3];
	jbp->A5.v[4] =  -6.0 * jbp->A0.v[4];
	jbp->A5.v[5] =  -6.0 * jbp->A0.v[5];
#endif
}

void setJointBlend2 (JOINT_BLEND_PARAMS *jbp, JNTS *j1, JNTS *j2, 
		double time, int numJ)
{
	/* Compute second part of Hermite correction polynomial */

	int j;

	for (j=0; j<numJ; j++)
	 { jbp->A1.v[j] = (j1->v[j] - j2->v[j] - jbp->A0.v[j]) / time;
	   jbp->A3.v[j] += -6.0 * jbp->A1.v[j];
	   jbp->A4.v[j] +=  8.0 * jbp->A1.v[j];
	   jbp->A5.v[j] += -3.0 * jbp->A1.v[j];
	 }
#if 0
	jbp->A1.v[0] = (j1->v[0] - j2->v[0] - jbp->A0.v[0]) / time;
	jbp->A1.v[1] = (j1->v[1] - j2->v[1] - jbp->A0.v[1]) / time;
	jbp->A1.v[2] = (j1->v[2] - j2->v[2] - jbp->A0.v[2]) / time;
	jbp->A1.v[3] = (j1->v[3] - j2->v[3] - jbp->A0.v[3]) / time;
	jbp->A1.v[4] = (j1->v[4] - j2->v[4] - jbp->A0.v[4]) / time;
	jbp->A1.v[5] = (j1->v[5] - j2->v[5] - jbp->A0.v[5]) / time;

	jbp->A3.v[0] += -6.0 * jbp->A1.v[0];
	jbp->A3.v[1] += -6.0 * jbp->A1.v[1];
	jbp->A3.v[2] += -6.0 * jbp->A1.v[2];
	jbp->A3.v[3] += -6.0 * jbp->A1.v[3];
	jbp->A3.v[4] += -6.0 * jbp->A1.v[4];
	jbp->A3.v[5] += -6.0 * jbp->A1.v[5];

	jbp->A4.v[0] +=  8.0 * jbp->A1.v[0];
	jbp->A4.v[1] +=  8.0 * jbp->A1.v[1];
	jbp->A4.v[2] +=  8.0 * jbp->A1.v[2];
	jbp->A4.v[3] +=  8.0 * jbp->A1.v[3];
	jbp->A4.v[4] +=  8.0 * jbp->A1.v[4];
	jbp->A4.v[5] +=  8.0 * jbp->A1.v[5];

	jbp->A5.v[0] += -3.0 * jbp->A1.v[0];
	jbp->A5.v[1] += -3.0 * jbp->A1.v[1];
	jbp->A5.v[2] += -3.0 * jbp->A1.v[2];
	jbp->A5.v[3] += -3.0 * jbp->A1.v[3];
	jbp->A5.v[4] += -3.0 * jbp->A1.v[4];
	jbp->A5.v[5] += -3.0 * jbp->A1.v[5];
#endif
}

void applyJointBlend (JNTS *result, JOINT_BLEND_PARAMS *jbp, JNTS *jnts,
		 double t, int numJ)
{
	/* Apply correction polynominal to input setpoint */

	float delj;
	int j;

	for (j=0; j<numJ; j++)
	 { delj = jbp->A0.v[j] + t*(jbp->A1.v[j] + t*(t*(jbp->A3.v[j] 
			       + t*(jbp->A4.v[j] + t*jbp->A5.v[j]))));
	   result->v[j] = jnts->v[j] + delj;
	 }
#if 0
	dJ.v[1] = jbp->A0.v[1] + t*(jbp->A1.v[1] + t*(t*(jbp->A3.v[1]
			       + t*(jbp->A4.v[1] + t*jbp->A5.v[1]))));
	dJ.v[2] = jbp->A0.v[2] + t*(jbp->A1.v[2] + t*(t*(jbp->A3.v[2]
			       + t*(jbp->A4.v[2] + t*jbp->A5.v[2]))));
	dJ.v[3] = jbp->A0.v[3] + t*(jbp->A1.v[3] + t*(t*(jbp->A3.v[3]
			       + t*(jbp->A4.v[3] + t*jbp->A5.v[3]))));
	dJ.v[4] = jbp->A0.v[4] + t*(jbp->A1.v[4] + t*(t*(jbp->A3.v[4]
			       + t*(jbp->A4.v[4] + t*jbp->A5.v[4]))));
	dJ.v[5] = jbp->A0.v[5] + t*(jbp->A1.v[5] + t*(t*(jbp->A3.v[5]
			       + t*(jbp->A4.v[5] + t*jbp->A5.v[5]))));
	
	result->v[0] = j->v[0] + dJ.v[0];
	result->v[1] = j->v[1] + dJ.v[1];
	result->v[2] = j->v[2] + dJ.v[2];
	result->v[3] = j->v[3] + dJ.v[3];
	result->v[4] = j->v[4] + dJ.v[4];
	result->v[5] = j->v[5] + dJ.v[5];
#endif
}
