#include <stdio.h>
#include "../h/grf4.h"
#include "../h/robotMath.h"
#include "robotsim.h"
#include <stdlib.h>
#include <ctype.h>
#include "../h/stringDefs.h"
#include "ringBuf.h"
#include "../h/cdefs.h"

typedef struct _plotRec {
	struct _plotRec *next;
	int code;
	double rng;	/* range */
	double frac;	/* frac of screen height corresponding to range */
	double off;	/* frac of screen height from bottom for base line */
} PlotRec;

extern int sampleInterval;

static PlotRec *plotRecs = NULL;
static int numPlotRecs = 0;
static char *codes = "123456";
static int tscaleAcc = 500*DEGTORAD;
static initDone = 0;
static RingBuf rbuf;
static float *fbuf;
static double plotOffset = 0;

#define RNG_SET		0x01
#define OFF_SET		0x02
#define FRAC_SET	0x04

#define MAX_TIME	30

static void jplotInitBuf (void)
{
	int size;
	void *buf;

	size = MAX_TIME*1000/sampleInterval;
	buf = malloc(sizeof(float)*MAXJNTS*size);
	if (buf == NULL)
	 { fprintf (stderr,
"jplotInitBuf() - couldn't allocate buffer\n");
	   exit (1);
	 }
	fbuf = (float*)buf;
	rbufInit (&rbuf, size, buf);
	initDone = 1;
}

static PlotRec *tailRec()
{
	PlotRec *rec;

	rec = plotRecs;
	while (rec && rec->next)
	 { rec = rec->next;
	 }
	return (rec);
}

static int jplotSpecParse (PlotRec *rec, char *str)
{
	int spec = 0;
	char *cp, *ep;

	cp = str;
	if (strchr (codes, *cp) == NULL)
	 { fprintf (stderr,
"jplotSpec(): unknown plot code %c in %s\n", *cp, str);
	   exit (1);
	 }
	rec->code = *cp;
	cp++;
	if (*cp == '\0')
	 { return spec;
	 }
	else if (*cp != ':')
	 { fprintf (stderr,
"jplotSpec(): ':' expected after plot code in %s\n", str);
	   exit (1);
	 }
	cp++;
	if (*cp != ':')
	 { rec->rng = strtod (cp, &ep);
	   if (cp == ep)
	    { fprintf (stderr,
"jplotSpec(): can't parse range spec in %s\n", str);
	      exit (1);
	    }
	   rec->rng *= DEGTORAD;
	   spec |= RNG_SET;
	   cp = ep;
	 }
	if (*cp == '\0')
	 { return spec;
	 }
	else if (*cp != ':')
	 { fprintf (stderr,
"jplotSpec(): ':' expected after range spec in %s\n", str);
	   exit (1);
	 }
	cp++;
	if (*cp != ':')
	 { rec->frac = strtod (cp, &ep);
	   if (cp == ep)
	    { fprintf (stderr,
"jplotSpec(): can't parse frac spec in %s\n", str);
	      exit (1);
	    }
	   spec |= FRAC_SET;
	   cp = ep;
	 }
	if (*cp == '\0')
	 { return spec;
	 }
	else if (*cp != ':')
	 { fprintf (stderr,
"jplotSpec(): ':' expected after frac spec in %s\n", str);
	   exit (1);
	 }
	cp++;
	rec->off = strtod (cp, &ep);
	if (cp == ep)
	 { fprintf (stderr,
"jplotSpec(): can't parse offset spec in %s\n", str);
	   exit (1);
	 }
	spec |= OFF_SET;
	return (spec);
}


void jplotSetSpec (char *str)
{
	PlotRec *rec;
	char *cp, *ep;
	int spec;

	rec = (PlotRec*)malloc(sizeof(PlotRec));
	rec->next = NULL;
	if (tailRec())
	 { tailRec()->next = rec;
	 }
	else
	 { plotRecs = rec;
	 }
	numPlotRecs++;
	spec = jplotSpecParse (rec, str);

	if (!(spec & RNG_SET))
	 { rec->rng = 180*DEGTORAD;
	 }
	if (!(spec & FRAC_SET))
	 { rec->frac = .2;
	 }
	if (!(spec & OFF_SET))
	 { rec->off = -1;
	 }
	printf ("rec->code = %c\n", rec->code);
	printf ("rec->frac = %g\n", rec->frac);
	printf ("rec->rng  = %g\n", RADTODEG*rec->rng);
}

void jplotSetTscaleAcc (double acc)
{
	tscaleAcc = acc;
}

void jplotSetOffset (double off)
{
	plotOffset = off;
}

void jplotStoreJnts (float *ang)
{
	static int wasMoving = 0;
	int moving;
	float *bufp;
	int nextTail;
	int bumpHead = 0;
	int doStore = 0;
	int j;

	if (numPlotRecs == 0)
	 { return;
	 }
	if (initDone == 0)
	 { jplotInitBuf();
	 }
	if (rbufNumItemsFree (&rbuf) == 0)
	 { bumpHead = 1;
	 }
	nextTail = rbufNextTail (&rbuf);
	bufp = fbuf+nextTail*MAXJNTS;
	if (rbufNumItemsQueued (&rbuf) <= 1)
	 { doStore = 1;
	   wasMoving = 0;
	 }
	else
	 { bufp = fbuf+rbuf.tail*MAXJNTS;
	   moving = 0;
	   for (j=0; j<MAXJNTS; j++)
	    { if (ABS(bufp[j]-ang[j]) > 0.01*DEGTORAD)
	       { moving = 1;
		 break;
	       }
	    }
	   if (moving || wasMoving)
	    { doStore = 1;
	    }
	   wasMoving = moving;
	 }
	if (doStore == 0)
	 { return;
	 }
	nextTail = rbufNextTail (&rbuf);
	bufp = fbuf+nextTail*MAXJNTS;	   
	for (j=0; j<MAXJNTS; j++)
	 { bufp[j] = ang[j];
	 }
	rbuf.tail = nextTail;
	if (bumpHead)
	 { rbuf.head = rbufNextHead (&rbuf);
	 }
}


static plotItem (PlotRec *rec, double off, double tscale, int nsamps,
		 double winw, double winh)
{
	float jval0, jval1, vel;
	double vscale;
	int idx, k;

	/* plot base line */

	grf4Color (GRF_BLUE);
	grf4Move2 (0, off);
	grf4Draw2 (winw, off);
	if (nsamps < 2)
	 { return;
	 }
	vscale = winh*rec->frac/rec->rng;
	grf4Color (GRF_GREEN);
	idx = rbufFromTail (&rbuf, nsamps);
	jval1 = fbuf[idx*MAXJNTS+rec->code-'1'];
	for (k=nsamps-1; k>=0; k--)
	 { jval0 = jval1;
	   idx = rbufFromTail (&rbuf, k);
	   jval1 = fbuf[idx*MAXJNTS+rec->code-'1'];
	   vel = (jval1-jval0)*1000/sampleInterval;

	   if (k==nsamps-1)
	    { grf4Move2 (0, vscale*vel+off);
	    }
	   else
	    { grf4Draw2 (tscale*(sampleInterval/1000.0)*(nsamps-1-k), 
			 vscale*vel+off);
	    }
	 }
}
	
void jplotVel (double winw, double winh)
{
	PlotRec *rec;
	double fracTotal;
	double tscale;
	int nsamps;	/* number of samples to plot */
	double off;

	if (numPlotRecs == 0)
	 { return;
	 }
	if (initDone == 0)
	 { jplotInitBuf();
	 }
	fracTotal = 0;
	for (rec=plotRecs; rec; rec=rec->next)
	 { fracTotal += rec->frac;
	 }
	/* compute tscale using tscaleAcc and the vscale of the
	   first item beging plotted */
	rec = plotRecs;
	tscale = tscaleAcc*(winh*rec->frac/rec->rng);
	nsamps = MIN((winw/tscale)*1000/sampleInterval+1,
		      rbufNumItemsQueued(&rbuf)-1);
	off = (2*fracTotal-rec->frac+plotOffset);
	for (rec=plotRecs; rec; rec=rec->next)
	 { plotItem (rec, off*winh, tscale, nsamps, winw, winh);
	   off -= 2*rec->frac;
	 }

}
