#ifndef _USEFUL_RINGBUF_H_
#define _USEFUL_RINGBUF_H_
 
/* Implements a ring-buffer structure. 


	|_|_|_|_|_|X|X|X|X|X|X|_|_|_|_|
		 ^	     ^
		 |	     |
		head        tail

	Items are added at the tail and removed from the head. 
	The buffer is empty when head == tail, so one slot is wasted.
	Tail is modified only by the producer, while head is modified
	only by the consumer, and this is done is such a way that the
	queue can be used by a separate producer and consumer task 
	without mutexes.

	To insert items:

		if (rbufNumItemsFree (rbuf) > 0)
		 { nextTail = rbufNextTail (rbuf);
		   buf[nextTail] = xxx;
		   rbuf->tail = nextTail;
		 }

	To remove items:

		if (rbufNumItemsQueued (rbuf) > 0)
		 { nextHead = rbufNextHead (rbuf);
		   xxx = buf[nextHead];
		   rbuf->head = nextHead;
		 }
*/

typedef struct {
	volatile int head;	/* index of the next item to be removed */
	volatile int tail;	/* index of the last item added */
	int size;		/* size of the queue. This is one more than
				   the number of items that can be queued */
	void *buf;		/* pointer to the buffer (convenience only) */
} RingBuf;

int rbufInit (RingBuf *rbuf, int size, void *buf);
int rbufNumItemsQueued (RingBuf *rbuf);
int rbufNumItemsFree (RingBuf *rbuf);
int rbufNextTail (RingBuf *rbuf);
int rbufNextHead (RingBuf *rbuf);
int rbufFromTail (RingBuf *rbuf, int k);
void rbufFlush (RingBuf *rbuf);

 
#endif
