/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd and Vincent Hayward.	    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*=~=~=~=~ McGill-Computer-Vision-and-Robotics-Laboratory--C-Module ~=~=~=~=
 *		
 * MODULE NAME:	stageJacob.c
 *
 * DESCRIPTION:	Jacobian computations for the Stanford arm.
 *		For description, see 'man pumaJacob'
 * 
 * AUTHOR/DATE:	John Lloyd, Spring 1987, after Vincent Hayward, Purdue, 1983
 *		Rewrote routines to interface with new RCI, winter 1988.
 *
 *~=~=~=~=~=~= McGill-Computer-Vision-and-Robotics-Laboratory ~=~=~=~=~=~=*/

#include <math.h>

#include "../h/fastmath.h"
#include "../h/cdefs.h"
#include "../h/robot_jls.h"
#include "../h/robotMath.h"
#include "../h/stage_jls.h"
#include "../h/stage_kynvar.h"
#include "../h/robotTypes.h"

#define EPSILONF 2.0e-6

#define J1	0
#define J2	1
#define J3	2

/* ---- Definition of var structure parameters ---- */

#define S3	svar->s3
#define C3	svar->c3

stageFwdJacobT(t, f, j6, kyn)
float *t;
FORCE *f;
float *j6;
KYN *kyn;
{
	STAGE_VAR var;

	stageUpdateVar (&var, j6, kyn);
	return (stageFwdJacobTVar (t, f, &var, kyn));
}

stageFwdJacob(dc, dj, j6, kyn)
DIFF *dc;
float *dj;
float *j6;
KYN *kyn;
{
	STAGE_VAR var;

	stageUpdateVar (&var, j6, kyn);
	return (stageFwdJacobVar (dc, dj, &var, kyn));
}

stageInvJacobT(f, t, j6, kyn, epsilon)
FORCE *f;
float *t;
float *j6;
KYN *kyn;
double epsilon;
{
	STAGE_VAR var;

	stageUpdateVar (&var, j6, kyn);
	return (stageInvJacobTVar (f, t, &var, kyn, epsilon));
}

stageInvJacob(dj, dc, j6, kyn, epsilon)
float *dj;
DIFF *dc;
float *j6;
KYN *kyn;
double epsilon;
{
	STAGE_VAR var;

	stageUpdateVar (&var, j6, kyn);
	return (stageInvJacobVar (dj, dc, &var, kyn, epsilon));
}

stageFwdJacobTVar(j, f, var, kyn) /*::*/
float *j;
FORCE *f;
void *var;
KYN *kyn;
{
	register STAGE_VAR *svar = (STAGE_VAR*)var;

	j[J1] = C3*f->f.x - S3*f->f.y;
	j[J2] = S3*f->f.x + C3*f->f.y;
	j[J3] = f->m.z;
}

stageFwdJacobVar(d, q, var, kyn) /*::*/
register DIFF_PTR d;
register float *q;
void *var;
KYN *kyn;
{
	register STAGE_VAR *svar = (STAGE_VAR*)var;

	d->t.x = C3*q[J1] + S3*q[J2];
	d->t.y =-S3*q[J1] + C3*q[J2];
	d->t.z = 0;
	d->r.x = 0;
	d->r.y = 0;
	d->r.z = q[J3];
}

stageInvJacobVar(q, d, var, kyn, epsilon)
register float *q;
register DIFF_PTR d;
void *var;
KYN *kyn;
double epsilon;
{
	STAGE_KYN *skyn = (STAGE_KYN*)kyn->ext;
	register STAGE_VAR *svar = (STAGE_VAR*)var;

	q[J1] = C3*d->t.x - S3*d->t.y;
	q[J2] = S3*d->t.x + C3*d->t.y;
	q[J3] = d->r.z;
	return (0);
}

stageInvJacobTVar(f, t, var, kyn, epsilon)
register FORCE *f;
register float *t;
void *var;
KYN *kyn;
double epsilon;
{
	STAGE_KYN *skyn = (STAGE_KYN*)kyn->ext;
	register STAGE_VAR *svar = (STAGE_VAR*)var;

	f->f.x = C3*t[J1] + S3*t[J2];
	f->f.y =-S3*t[J1] + C3*t[J2];
	f->f.z = 0;
	f->m.x = 0;
	f->m.y = 0;
	f->m.z = t[J3];
	return (0);
}

/* Jacobian updating functions */

stageUpdateVar (var, ang, kyn)
void *var;
register float *ang;
KYN *kyn;
{
	register STAGE_VAR *svar = (STAGE_VAR*)var;

	SINCOS (ang[J3], &S3, &C3);
	stageUpdateVarXsincos (var, ang, kyn);
}

stageUpdateVarXsincos (var, ang, kyn)
void *var;
float *ang;
KYN *kyn;
{
   
	register STAGE_KYN *skyn = (STAGE_KYN*)kyn->ext;
	register STAGE_VAR *svar = (STAGE_VAR*)var;

	svar->d[J1] = ang[J1];
	svar->d[J2] = ang[J2];
}
