/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd and Vincent Hayward.	    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*=~=~=~=~ McGill-Computer-Vision-and-Robotics-Laboratory--C-Module ~=~=~=~=
 *		
 * MODULE NAME:	stageKinematic.c
 *
 * DESCRIPTION:	Forward and inverse kinematic routines for the STAGE robots.
 * 
 * AUTHOR/DATE:	John Lloyd, Spring 1987, after Vincent Hayward, Purdue, 1983
 *		Rewrote routines to interface with new RCI, winter 1988.
 *
 *~=~=~=~=~=~= McGill-Computer-Vision-and-Robotics-Laboratory ~=~=~=~=~=~=*/

#include <math.h>

#include "../h/fastmath.h"
#include "../h/cdefs.h"
#include "../h/robot_jls.h"
#include "../h/robotMath.h"
#include "../h/stage_jls.h"
#include "../h/stage_kynvar.h"
#include "../h/robotTypes.h"

#include "../h/rcclDefs.h"	/* for F_UNDEF */

#define EPSILONF 2.0e-6

/* performs mod 2 pi */

static float range(a) /*##*/	/* Sets a to  0 <= a < pi2  */
double a;
{
	if (a > 15. || a < -15.)
	 { return(10.);
	 }

	while(a >= PIT2)
	 { a -= PIT2;
	 }
	while(a < 0.)
	 { a += PIT2;
	 }
	if (a > PIT2)
	 { return(0.);
	 }
	return(a);
}

/* ---- Definition of link parameters to make code more readable ---- */

#define J1	0
#define J2	1
#define J3	2

stageFwdKinematics (t, conf, j, kyn, var)
TRSF *t;	  	  /* resulting transform */
unsigned long *conf;	  /* returns configuration information */
float *j;		  /* joints in range coordinates + configuration */
KYN *kyn;	  	  /* kinematic parameters */
void *var;		  /* palce for storing computing sines/cosines */
{
	register STAGE_VAR *svar = (STAGE_VAR*)var;
	GEN_KYN *gen = kyn->gen;

	/* sines and cosines */

	float c3, s3;

	SINCOS (j[J3], &s3, &c3);

	t->n.x = c3;
	t->n.y = s3;
	t->n.z = 0;

	t->o.x = -s3;
	t->o.y = c3;
	t->o.z = 0;

	t->a.x = 0;
	t->a.y = 0;
	t->a.z = 1;

	t->p.x = j[J1];
	t->p.y = j[J2];
	t->p.z = 0;

	if (conf != NULL)
	 { *conf = 0;
	 }

	if (var != NULL)	/* Update sines and cosinesd if needed */
	 { svar->s3 = s3;
	   svar->c3 = c3;
	   svar->d[J1] = j[J1];
	   svar->d[J2] = j[J2];
	 }
}

stageInvKinematics(j, t, conf, refj, kyn, jls, var) /*::*/
float *j;	  	  /* joints in range coordinates + configuration */
TRSF *t;	  /* input transform 			  	 */
unsigned long conf;	  /* configuration information			 */
float *refj;		  /* reference joints				 */
KYN *kyn;		  /* kinematic parameters			 */
void *var;		  /* run-time variables 			 */
JLS *jls;		  /* joint level data -- usd to get min. joints	 */
{
	STAGE_KYN *skyn = (STAGE_KYN*)kyn->ext;
	register STAGE_VAR *svar = (STAGE_VAR*)var;
	GEN_KYN *gen = kyn->gen;
	TRSF TR;
	int code = 0;
	VECT u;
	double ca, sa;

	float s3, c3;
	float ang[STAGEJ];			/* Non-ranged angle values */

	ang[J1] = t->p.x;
	ang[J2] = t->p.y;

	TR.n = t->n;
	TR.o = t->a;
	TR.a = t->a;
	zeroVect (&TR.p);

	crossVect (&u, zunit, &TR.a);
	sa = norm2Vect (&u);
	if (sa > EPSILONF)
	 { TRSF TC;
	   ca = TR.a.z;
	   rotToTrsf (&TC, &u, -atan2(sa, ca));
	   multTrsf (&TR, &TC, &TR);
	 }
	ang[J3] = atan2 (TR.n.y, TR.n.x);
	code |= stageRangeAngles (j, ang, refj, kyn, jls);

	if (var != NULL)	/* Update sines and cosines if needed */
	 { svar->s3 = s3;
	   svar->c3 = c3;
	   svar->d[J1] = j[J1];
	   svar->d[J2] = j[J2];
	 }
	return(code);
}

stageRangeAngles (r, a, refj, kyn, jls)
register float *r, *a;
float *refj;
KYN *kyn;
JLS *jls;
{
	register float *jmin, *jmax;
	register double angle;
	register double refAngle;
	int code = 0;

	jmin = jls->jmin;
	jmax = jls->jmax;

	if (a[J1] <= jmin[J1] || a[J1] >= jmax[J1])
	 { code |= (1 << 0);
	 }
	r[J1] = a[J1];

	if (a[J2] <= jmin[J2] || a[J2] >= jmax[J2])
	 { code |= (1 << 1);
	 }
	r[J2] = a[J2];

	/* Range of joint 3 is potentially unlimited, so we need to
	   use reference angles */

	angle = a[J3];
	refAngle = (refj ? refj[J3] : (jmax[J3] + jmin[J3])/2);
	if (angle <= refAngle)
	 { while (angle <= refAngle)	
	    { angle += PIT2;
	    }
	   if (angle - refAngle <= PI && angle < jmax[J3])
	    { r[J3] = angle;		
	    }
	   else
	    { angle -= PIT2;
	      if ((r[J3] = angle) <= jmin[J3])
	       { code |= (1 << 5);
	       }
	    }
	 }
	else
	 { while (angle > refAngle)
	    { angle -= PIT2;
	    }
	   if (refAngle - angle <= PI && angle > jmin[J3])
	    { r[J3] = angle;
	    }
	   else
	    { angle += PIT2;
	      if ((r[J3] = angle) >= jmax[J3])
	       { code |= (1 << J3);
	       }
	    }
	 }
	r[J3] = angle;
	return (code);
}

stageSolveConf(conf, j, kyn)
unsigned long *conf;	/* configuration description */
register float *j;	/* joints in radians */
KYN *kyn;
{
	*conf = 0;
}

