/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*=~=~=~=~ McGill-Computer-Vision-and-Robotics-Laboratory--C-Module ~=~=~=~=
 *		
 * NAME:	testJacobian
 *
 * USAGE:	testJacobian <robotName> [-i]
 *
 * DESCRIPTION: performs a test of the jacobian functions for the arm.
 *		The tests are done by playing the functions off against
 *		each other: a function is applied to a set of data,
 *		then the inverse is applied, and the final output is
 *		compared to the original input.
 * 
 *		fwdJacob() is played off against invJacob(), and
 *		fwdJacobT() is played off against invJacobT().
 *
 *		The program performs about 2000 tests. These include 
 *		cases away from, close to, and at singularities. The results
 *		printed out describe the maximum observed percent error
 *		between the input and output values.
 *
 *		The testing AT SINGULARITIES is a bit crude, but the best
 *		I could do quickly: The functions are run through once in
 *		in an attempt to zero the values associated with degrees
 *		of freedom which have been lost; running the data through
 *		again should produce meaningfull results.
 *
 * ARGUMENTS:	The -i option allows the user to interactively run 
 *		individual test on specific sets of joints angles.
 *	
 * NOTES:	I wrote this because I was always playing with the kinematic
 *		functions, but was never sure I hadn't screwed them up in the
 *		process.
 *
 * AUTHOR/DATE: John Lloyd, Spring 87
 *
 * UPDATES:
 *
 *~=~=~=~=~=~= McGill-Computer-Vision-and-Robotics-Laboratory ~=~=~=~=~=~=*/

#include <stdio.h>
#include <math.h>

#include "../h/cdefs.h"
#include "../h/robot_jls.h"
#include "../h/robotMath.h"
#include "../h/stage_kynvar.h"
#include "../h/vector.h"
#include "../h/robotTypes.h"
#include "../h/fatalMsg.h"

#define EPSILONF	2.0e-6

#define	COMMAND(s)	(strcmp (key_buf, s) == 0)
#define	COMMANDS(s)	(strncmp (key_buf, s, strlen(s)) == 0)
char key_buf [256];

KYN *kyn;
void *var;
JLS *jls;

main(argc, argv)
int argc;
char **argv;
{
	char *robotName = NULL;
	int interactive = 0;

	argc--; argv++;
	while (argc)
	 { if (ARGUMENT ("-i"))
	    { interactive = 1;
	    }
	   else if (robotName == NULL)
	    { robotName = *argv;
	    }
	   else
	    { robotName = NULL;
	    }
	   argc--; argv++;
	 }
	if (robotName == NULL)
	 { usageMsg (-1, "testJacobian <robotName> [-i]\n");
	 }

	if ((jls = getJls(robotName)) == NULL)
	 { printErrors();
	   exit (-1);
	 }
	if ((kyn = getKyn(robotName, NULL)) == NULL)
	 { printErrors();
	   exit (-1);
	 }
	var = (void*)malloc(getVarSize(robotName));

	if (interactive)
	 { char *keytree, *tree_match_parse(), *tree_match_err_at();
	   float epsilon;

	   float angles[STAGEJ];

	   keytree = tree_match_parse("					\
(	   quit								\
	   set 								\
	    (angles							\
	     epsilon							\
	      (%f", &epsilon, ",0))					\
	   show								\
	   fwdJacobT							\
	   fwdJacob							\
	   invJacobT							\
	   invJacob							\
)",0);

	   if (keytree == 0)
	    { printf ("ERROR/Parse error at %.30s\n", tree_match_err_at());
	      exit (-1);
	    }

	   stageUpdateVar (var, angles, kyn);

	   do
	    { tree_match (keytree, "TEST JACOBIAN> ", key_buf);

	      if (COMMAND ("set angles"))
	       { 
		 printf ("5 angles, in degrees > ");
		 if (scanVf (angles, jls->numJ) != 1)
		  { printf ("ERROR reading values\n");
		    continue;
		  }
		 degreeToRadian (angles, angles, jls);

		 stageUpdateVar (var, angles, kyn);
	       }
	      else if (COMMAND ("show"))
	       { float ang[STAGEJ];
		 printf ("epsilon = %g\n", epsilon);
		 radianToDegree (ang, angles, jls);
		 printVf ("JOINT ANGLES: %8.3f\n", ang, jls->numJ);
	       }
	      else if (COMMAND ("fwdJacob"))
	       { float dtheta1[STAGEJ], dtheta2[STAGEJ];
		 DIFF dc1, dc2;
		 int code;

		 printf ("%d delta theta values > ", jls->numJ);
		 if (scanVf (dtheta1, jls->numJ) != 1)
		  { printf ("ERROR reading values\n");
		    continue;
		  }
		 stageFwdJacobVar (&dc1, dtheta1, var, kyn);

		 printf ("Cartesian deltas from fwdJacob():\n");
		 printVf ("\t%10.5f\n", (float*)&dc1, 6);

		 code = stageInvJacobVar(dtheta2, &dc1, var, kyn, epsilon);

		 printf ("Joint deltas from invJacob(), status %x:\n", code);
		 printVf ("\t%10.5f\n", dtheta2, jls->numJ);

		 stageFwdJacobVar (&dc2, dtheta2, var, kyn);

		 printf ("Cartesian deltas from fwdJacob():\n");
		 printVf ("\t%10.5f\n", (float*)&dc2, 6);
	       }		 
	      else if (COMMAND ("invJacob"))
	       { float dtheta1[STAGEJ], dtheta2[STAGEJ];
		 DIFF dc1, dc2;
		 int code;

		 printf ("6 Cartesian delta values > ");
		 if (scanVf ((float*)&dc1, 6) != 1)
		  { printf ("ERROR reading values\n");
		    continue;
		  }
		 code = stageInvJacobVar(dtheta1, &dc1, var, kyn, epsilon);

		 printf ("Joint deltas from invJacob(), status %x:\n", code);
		 printVf ("\t%10.5f\n", dtheta1, jls->numJ);

		 stageFwdJacobVar (&dc2, dtheta1, var, kyn);

		 printf ("Cartesian deltas from fwdJacob():\n");
		 printVf ("\t%10.5f\n", (float*)&dc2, 6);

		 code = stageInvJacobVar(dtheta2, &dc2, var, kyn, epsilon);

		 printf ("Joint deltas from invJacob(), status %x:\n", code);
		 printVf ("\t%10.5f\n", dtheta2, jls->numJ);
	       }		 
	      else if (COMMAND ("fwdJacobT"))
	       { float torque1[STAGEJ], torque2[STAGEJ];
		 FORCE f1, f2;
		 int code;

		 printf ("6 force values > ");
		 if (scanVf ((float*)&f1, 6) != 1)
		  { printf ("ERROR reading values\n");
		    continue;
		  }
		 stageFwdJacobTVar (torque1, &f1, var, kyn);

		 printf ("Torque values from fwdJacobT():\n");
		 printVf ("\t%10.5f\n", torque1, jls->numJ);

		 code = stageInvJacobTVar(&f2, torque1, var, kyn, epsilon);

		 printf ("Force values from invJacobT(), status %x:\n", code);
		 printVf ("\t%10.5f\n", (float*)&f2, 6);

		 stageFwdJacobTVar (torque2, &f2, var, kyn);

		 printf ("Torque values from fwdJacobT():\n");
		 printVf ("\t%10.5f\n", torque2, jls->numJ);
	       }		 
	      else if (COMMAND ("invJacobT"))
	       { float torque1[STAGEJ], torque2[STAGEJ];
		 FORCE f1, f2;
		 int code;

		 printf ("%d torque values > ", jls->numJ);
		 if (scanVf (torque1, jls->numJ) != 1)
		  { printf ("ERROR reading values\n");
		    continue;
		  }
		 code = stageInvJacobTVar(&f1, torque1, var, kyn, epsilon);

		 printf ("Force values from invJacobT(), status %x:\n", code);
		 printVf ("\t%10.5f\n", (float*)&f1, 6);

		 stageFwdJacobTVar (torque2, &f1, var, kyn);

		 printf ("Torque values from fwdJacobT():\n");
		 printVf ("\t%10.5f\n", torque2, jls->numJ);

		 code = stageInvJacobTVar(&f2, torque2, var, kyn, epsilon);

		 printf ("Force values from invJacobT(), status %x:\n", code);
		 printVf ("\t%10.5f\n", (float*)&f2, 6);
	       }		 
	    }
	   while (!COMMAND ("quit"));
	 }
	else	/* run automatric checking sequence */
	 { 
	   int numchecks, i, code;
	   float angles1[6];
	   float angles2[6];
	   float maxAnglesD[6];
	   float maxAnglesT[6];
	   float normD, normT, maxnormD, maxnormT;
	   int numsing;

#define END		-1.0e30

	    { static float testset[] = 
	       {  0.0, 120.0, -70.0, END,
		-30.0,  60.0, 120.0, END,
		250.0, 500.0, END,
		  0.0, 120.0, END,
		 45.0,   0.0, -45.0, END,
		120.0,-120.0,   0.0, END };

	      numchecks = testAnglesInit (testset, jls);
printf ("\nTesting %d positions ...\n\n", numchecks);
	      maxnormD = maxnormT = -1.0;
	      numsing = 0;
	      while (testAnglesGet (angles1))
	       { 
		 if (code = test_jacobian (&normD, &normT, angles1, EPSILONF))
		  { numsing++;
		  }
		 if (normD > maxnormD)
	          { maxnormD = normD;
		    copyVf (maxAnglesD, angles1, jls->numJ);
		  }
		 if (normT > maxnormT)
	          { maxnormT = normT;
		    copyVf (maxAnglesT, angles1, jls->numJ);
		  }
	       }
/*printf ("%d singular positions detected\n", numsing);*/
printf ("Maximum %% error for fwdJacob() <-> invJacob() is %g at angles \n", 
	maxnormD);
printa (" ",maxAnglesD);
printf ("Maximum %% error for fwdJacobT() <-> invJacobT() is %g at angles \n", 
	maxnormT);
printa (" ",maxAnglesT);
	    }

	    { static float testset[] = 
		{  0.0, 120.0, END,
		  60.0, -60.0, END,
		 250.0, 500.0, END,
		  20.0, 120.0, END,
		 -45.0,  45.0, END,
		 120.0,-120.0, END };

	      int i;
	      float distance;

	      numchecks = testAnglesInit (testset, jls);
printf ("\nTesting %d more positions ...\n\n", numchecks);
	      maxnormD = maxnormT = -1.0;
	      numsing = 0;
	      while (testAnglesGet (angles1))
	       { 
		 if (code = test_jacobian (&normD, &normT, 
					   angles2, EPSILONF))
		  { numsing++;
		  }
		 if (normD > maxnormD)
		  { maxnormD = normD;
		    copyVf (maxAnglesD, angles2, jls->numJ);
		  }
		 if (normT > maxnormT)
		  { maxnormT = normT;
		    copyVf (maxAnglesT, angles2, jls->numJ);
		  }
	       }
/*printf ("%d singular positions detected\n", numsing);*/
printf ("Maximum %% error for fwdJacob() <-> invJacob() is %g at angles \n", 
	maxnormD);
printa (" ",maxAnglesD);
printf ("Maximum %% error for fwdJacobT() <-> invJacobT() is %g at angles \n", 
	maxnormT);
printa (" ",maxAnglesT);
	    }
	 }
}

test_jacobian (normD, normT, angles, epsilon)
float *normD, *normT;
float *angles;
float epsilon;
{
	int codeD, codeT;

	DIFF dc, dc_ref;
	FORCE f, f_ref;
	float dtheta[STAGEJ], dtheta_ref[STAGEJ];
	float torque[STAGEJ], torque_ref[STAGEJ];

	static DIFF dc_test = {1.0, 2.0, 3.0, 4.0, 5.0, 6.0 };
	static FORCE f_test = {1.0, 2.0, 3.0, 4.0, 5.0, 6.0 };
	static float dtheta_test[STAGEJ] = {1.0, 2.0, 3.0 };
	static float torque_test[STAGEJ] = {1.0, 2.0, 3.0 };

	stageUpdateVar (var, angles, kyn);

	copyVf (dtheta_ref, dtheta_test, jls->numJ);
	stageFwdJacobVar (&dc, dtheta_ref, var, kyn);
	codeD = stageInvJacobVar (dtheta, &dc, var, kyn, epsilon);

	if (normVf (dtheta, jls->numJ) > normVf(dtheta_ref, jls->numJ)*1.01 )
	 { 
	   printf ("Input and reconstructed joint deltas differ greatly,\n");
	   printf ("(norm %g vs. %g), at angles\n",
		   normVf(dtheta, jls->numJ), normVf(dtheta_ref,jls->numJ));
	   printa (" ", angles);
	 }

	subtractVf (dtheta, dtheta, dtheta_ref, jls->numJ);
	*normD = normVf (dtheta, jls->numJ) / normVf (dtheta_ref, jls->numJ);

       	copyVf (torque_ref, torque_test, jls->numJ);
	codeT = stageInvJacobTVar (&f, torque_ref, var, kyn, epsilon);
	stageFwdJacobTVar (torque, &f, var, kyn);

	if (normVf(torque, jls->numJ) > normVf(torque_ref, jls->numJ)*1.01)
	 { 
	   printf ("Input and reconstructed torque values differ greatly,\n");
	   printf ("(norm %g vs. %g), at angles\n",
		   normVf(torque,jls->numJ), normVf(torque_ref,jls->numJ));
	   printa (" ", angles);
	 }

	subtractVf (torque, torque, torque_ref, jls->numJ);
	*normT = normVf (torque, jls->numJ) / normVf (torque_ref, jls->numJ);

	if (codeD != codeT)
	 { 
printf ("invJacob() and invJacobT() find singularites %x and %x, at angles\n",
	codeD, codeT);
printa (" ", angles);
	 }
	return (codeD | codeT);
}

	 
printa (s, fp)
char *s;
float *fp;
{
	char buf[256];
	float a[6];

	strcpy (buf, s);
	strcat (buf, " %10.5lf\n");
	radianToDegree (a, fp, jls);
	printVf (buf, a, jls->numJ);
}
