/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*=~=~=~=~ McGill-Computer-Vision-and-Robotics-Laboratory--C-Module ~=~=~=~=
 *		
 * NAME:	testKinematics <robotName>
 *
 * USAGE:	testKinematics <robotName> [-i]
 *
 * DESCRIPTION: performs a test of the forward and inverse kinematic 
 *		functions of the arm. It is done simply by playing them
 *		off against each other ... the program runs through a 
 *		sequence of joint angles, computes the configuration,
 *		the forward kinematics, and then computes the inverse
 *		kinematics.
 *
 *		The program performs about 2000 tests. These include 
 *		cases away from and close to singularities. The results
 *		printed out describe the maximum observed norm of the
 *		difference (in degrees) between the original angles and the
 *		ones recomputed by the inverse kinematics.
 *
 * ARGUMENTS:	The -i option allows the user to interactively run tests
 *		on specific sets of joints angles.
 *
 * FILES:
 *	
 * NOTES:	I wrote this because I was always playing with the kinematic
 *		functions, but was never sure I hadn't screwed them up in the
 *		process.
 *
 * AUTHOR/DATE: John Lloyd, Spring 87
 *
 * UPDATES:
 *
 *~=~=~=~=~=~= McGill-Computer-Vision-and-Robotics-Laboratory ~=~=~=~=~=~=*/

#include <stdio.h>
#include <math.h>

#ifdef Vxw_TARGET
#define main testKinematics_
#endif

#include "../h/cdefs.h"
#include "../h/robot_jls.h"
#include "../h/robotMath.h"
#include "../h/stage_kynvar.h"
#include "../h/vector.h"
#include "../h/robotTypes.h"
#include "../h/cputime.h"
#include "../h/fatalMsg.h"

#define	COMMAND(s)	(strcmp (key_buf, s) == 0)
#define	COMMANDS(s)	(strncmp (key_buf, s, strlen(s)) == 0)
char key_buf [256];

KYN *kyn;
void *var;
JLS *jls;

double range2Pi(a)
double a;
{
	/* return a in the range [0,2 Pi) */

	double x;

	if (a >= 0.0)
	 { x = a - 2*PI*(int)(a/(2*PI));
	 }
	else
	 { x = a - 2*PI*((int)(a/(2*PI))-1);
	 }
	if (x < 0.0 || x >= 2*PI)
	 { x = 0.0;
	 }
	return (x);	
}

main(argc, argv)
int argc;
char **argv;
{
	int icount;
	char interactive = 0;
	char *robotName = NULL;

	argc--; argv++;
	while (argc)
	 { if (ARGUMENT ("-i"))
	    { interactive = 1;
	    }
	   else if (robotName == NULL)
	    { robotName = *argv;
	    }
	   else
	    { robotName = NULL;
	      break;
	    }
	   argc--; argv++;
	 }
	if (robotName == NULL)
	 { usageMsg (-1, "testKinematics <robotName> [-i]\n");
	 }

	if ((jls = getJls(robotName)) == NULL)
	 { printErrors();
	   exit (-1);
	 }
	if ((kyn = getKyn(robotName, NULL)) == NULL)
	 { printErrors();
	   exit (-1);
	 }
	var = (void*)malloc(getVarSize(robotName));

	if (interactive)
	 { char *keytree, *tree_match_parse(), *tree_match_err_at();
	   float angles[STAGEJ];
	   float result[STAGEJ];
	   TRSF T6;

	   keytree = tree_match_parse("					\
(	   quit								\
	   set 								\
	    (angles)							\
	   solve							\
	   speed							\
	    (%d", &icount, ",,,	\"iteration count\")			\
)",0);

	   if (keytree == 0)
	    { printf ("ERROR/Parse error at %.30s\n", tree_match_err_at());
	      exit (-1);
	    }
	   do
	    { int status;

	      tree_match (keytree, "TEST KINEMATICS> ", key_buf);

	      if (COMMAND ("set angles"))
	       { 
	         printf ("Enter j6 in degrees> ");
		 if (scanVf (angles, jls->numJ) != 1)
	          { fprintf (stderr, "ERROR/ cannot read values\n");
		    continue;
	          }
		 degreeToRadian (angles, angles, jls);
	       }
	      else if (COMMAND ("solve"))
	       { unsigned long conf, testConf;
		 char cbuf[4];

		 stageFwdKinematics (&T6, &conf, angles, kyn, NULL);
		 stageSolveConf (&testConf, angles, kyn);
		 status = stageInvKinematics (result, &T6, conf, angles, 
							   kyn, jls, NULL);
		 printf ("FwdKin conf is '%s' (0x%x)\n", 
			 configToStr(cbuf, conf, kyn), conf);
		 printf ("SolveConf conf is '%s' (0x%x)\n", 
			 configToStr(cbuf, testConf, kyn), testConf);
		 printf ("IK status = 0x%x\n", status);
		 radianToDegree (result, result, jls);
		 printTrsf ("T6:    %m\n", &T6);
		 printVf ("Result: %10.4f\n", result, jls->numJ);
	       }
	      else if (COMMAND ("speed"))
	       { float loopTime;
		 float time1;
		 float time2;
		 unsigned long conf;
		 int i;

		 /* cputimeInit(); */

		 time1 = cputime(User);
		 for (i=0; i<icount; i++)
		    ;
		 time2 = cputime(User);
		 loopTime = time2-time1;

		 time1 = cputime(User);
		 for (i=0; i<icount; i++)
		  { stageFwdKinematics (&T6, &conf, jls->bestAngles, kyn, var);
		  }
		 time2 = cputime(User);
		 
		 printf ("forward kinematics: %g usec.\n",
			  (time2-time1-loopTime)*1e6/icount);

		 time1 = cputime(User);
		 for (i=0; i<icount; i++)
		  { stageInvKinematics (result, &T6, conf, jls->bestAngles, 
					       kyn, jls, var);
		  }
		 time2 = cputime(User);
		 
		 printf ("inverse kinematics: %g usec.\n",
			  (time2-time1-loopTime)*1e6/icount);

		 time1 = cputime(User);
		 for (i=0; i<icount; i++)
		  { stageInvKinematics (result, &T6, conf, jls->bestAngles, 
					       kyn, jls, 0);
		  }
		 time2 = cputime(User);

		 /* cputimeExit(); */
		 
		 printf ("inverse kinematics (no var update): %g usec.\n",
			  (time2-time1-loopTime)*1e6/icount);
	       }
	    }
	   while (!COMMAND ("quit"));
	 }
	else
	 {
	   int numtests, singular, numsing;
	   float angles1[6];
	   float angles2[6];
	   float norm, worstNorm;
	   int errj, maxErrJ;
	   float worstAngles[6];
	   float *testset1;
	   float *testset2;

#define END		-1.0e30

	   static float newtestset1[] = 
	    { 100.0, 0.0, -100.0, END,
	      200.0, 0.0, -200.0, END,
		5.0, 500.0, END,
	       0.0, 120.0, END,
	      45.0,  20.0, -60.0, END };

	   static float newtestset2[] =
	     {  0.0, 120.0, END,
	       200.0, -200.0, END,
	      250.0, 500.0, END,
	       20.0, 120.0, END,
	      -45.0,  0.0, END };

	   testset1 = newtestset1;
	   testset2 = newtestset2;

	   /* Testing away from singularities */

	   numtests = testAnglesInit (testset1, jls);
	   printf (
"\nTesting with %d positions ...\n\n", numtests);
	   worstNorm = -1.0;
	   maxErrJ = 0;
	   numsing = 0;
	   while (testAnglesGet (angles1))
	    { 
	      if (test_solut (&norm, &errj, angles1))
	       { numsing++;
	       }
	      if (norm > worstNorm)
	       { worstNorm = norm;
		 maxErrJ = errj;
		 copyVf (worstAngles, angles1, jls->numJ);
	       }
	    }
	   printf ("Worst average percent error = %g for angle %d\n", 
		    100*worstNorm, maxErrJ+1);
	   printa (" ", worstAngles);

	   /* Test near singularities */

	   numtests = testAnglesInit (testset2, jls);
	   printf (
"\nTesting with %d more positions ... \n\n", numtests);
	   worstNorm = -1.0;
	   numsing = 0;
	   while (testAnglesGet (angles1))
	    { if (test_solut (&norm, &errj, angles2))
	       { numsing++;
	       }
	      if (norm > worstNorm)
	       { worstNorm = norm;
	         maxErrJ = errj;
	         copyVf (worstAngles, angles2, jls->numJ);
   	       }
	    }
	   printf ("Worst average percent error = %g for angle %d\n",
		    100*worstNorm, maxErrJ+1);
	   printa (" ", worstAngles);
	 }
}

test_solut (norm, errj, angles1)
float *norm;
int *errj;
float *angles1;
{	
	TRSF T6;
	float j60[STAGEJ];
	float j61[STAGEJ];
	float j62[STAGEJ];
	float angles2[STAGEJ];
	int code = 0;
	unsigned long conf1;
	unsigned long conf2;
	int j;
	float err;

	stageFwdKinematics (&T6, &conf1, angles1, kyn, NULL);
	stageSolveConf (&conf2, angles1, kyn);
	if (conf1 != conf2)
	 { printf (
"solveConf() and fwdKinematics() compute configuartions 0x%x and 0x%x\n",
	           conf2, conf1);
	   printa ("at angles ", angles1);
	 }
	code = stageInvKinematics (angles2, &T6, conf1, angles1, 
							kyn, jls, NULL);
	*norm = err = 0.0;
	*errj = 0;
	for (j=0; j<jls->numJ; j++)
	 { err = FABS(angles2[j]-angles1[j]);
	   if (jls->jtype[j] == 'R')
	    { if ((err = range2Pi(err)) > PI)
	       { err = FABS(err - 2*PI);
	       }
	      err /= PI;
	    }
	   else
	    { err /= (jls->jmax[j]-jls->jmin[j]);
	    }
	   if (err > *norm)
	    { *norm = err;
	      *errj = j;
	    }
	 }
	return (code);
}

printa (s, fp)
char *s;
float *fp;
{
	char buf[256];
	float a[6];

	strcpy (buf, s);
	strcat (buf, " %10.5lf\n");
	radianToDegree (a, fp, jls);
	printVf (buf, a, jls->numJ);
}
