/* A set of routines to provide some simple system-independent control
   over terminal input lines */

#include "../h/termline.h"

termlineGet (fd, props)
int fd;
TermProps *props;
{
	int status;
#ifdef HAS_TERMIOS
	status = tcgetattr (fd, &props->data);
#else
#ifdef HAS_TERMIO
	status = ioctl (fd, TCGETA, &props->data);
#else
#ifdef HAS_SGTTY
	status = ioctl (fd, TIOCGETP, &props->data);
#else
#ifdef Vxw_TARGET
	status = ioctl (fd, FIOGETOPTIONS);
#else
	status = -1;
#endif   
#endif
#endif
#endif
	return (status);
}

termlineSet (fd, props)
int fd;
TermProps *props;
{
	int status;
#ifdef HAS_TERMIOS
	status = tcsetattr (fd, TCSANOW, &props->data);
#else
#ifdef HAS_TERMIO
	status = ioctl (fd, TCSETA, &props->data);
#else
#ifdef HAS_SGTTY
	status = ioctl (fd, TIOCSETP, &props->data);
#else
#ifdef Vxw_TARGET
	status = ioctl (fd, FIOSETOPTIONS, props->data);
#else
	status = -1;
#endif
#endif
#endif
#endif
	return (status);
}

termlineOutputFlow (fd, on)
int fd, on;
{
	int status = 1;
#ifdef HAS_TERMIOS
	status = tcflow (fd, on ? TCOON : TCOOFF);
#else
#ifdef HAS_TERMIO
	status = ioctl (fd, TCXONC, on ? 1 : 0);
#else
#ifdef HAS_SGTTY
	status = ioctl (fd, on ? TIOCSTART : TIOCSTOP, 0);
#else
	status = -1;
#endif
#endif
#endif
	return (status);
}

termlineRawMode (props)
TermProps *props;
{
	int status = 1;
#if defined(HAS_TERMIOS) || defined(HAS_TERMIO)
#ifdef IXANY
	props->data.c_iflag &= (IXON|IXANY);
#else
	props->data.c_iflag &= (IXON);
#endif
	props->data.c_oflag &= ~(OPOST);
	/* seems like clearing ISIG is redundant since we are clearing
	   ICANON, but the Solaris ttcompat(7) document indiactes we
	   should do this. */
	props->data.c_lflag &= ~(ISIG|ICANON);
#ifdef XCASE
	props->data.c_lflag &= ~(XCASE);
#endif
	props->data.c_cflag &= ~(PARENB|PARODD|CSIZE);
	props->data.c_cflag |=  (CS8);
	props->data.c_cc[VMIN] = 1;
	props->data.c_cc[VTIME] = 0;
#else
#ifdef HAS_SGTTY
	props->data.sg_flags |= RAW;	/*  raw mode (passall) */
	props->data.sg_flags &= ~ECHO;	/*  no echo */
#else
	status = -1;
#endif
#endif
	return (status);
}

termlineCbreakMode (props)
TermProps *props;
{
	int status = 1;
#if defined(HAS_TERMIOS) || defined(HAS_TERMIO)
	props->data.c_lflag &= ~ICANON;
	props->data.c_cc[VMIN] = 1;
	props->data.c_cc[VTIME] = 0;
#else
#ifdef HAS_SGTTY
	props->data.sg_flags |= CBREAK;
#else
#ifdef  Vxw_TARGET
	props->data &= ~(OPT_TANDEM|OPT_LINE);
#else
	status = -1;
#endif
#endif
#endif
	return (status);
}

termlineEchoOff (props)
TermProps *props;
{
	int status = 1;
#if defined(HAS_TERMIOS) || defined(HAS_TERMIO)
	/* some of these echo bits don't really need to be cleared,
	   because clearing ECHO would disable them anyway, but the
	   documentation is a bit fuzzy on the details so we clear 
	   them all */
	props->data.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL);
#ifdef ECHOCTL
	props->data.c_lflag &= ~(ECHOCTL);
#endif
#ifdef ECHOPRT
	props->data.c_lflag &= ~(ECHOPRT);
#endif
#ifdef ECHOKE
	props->data.c_lflag &= ~(ECHOKE);
#endif
#else
#ifdef HAS_SGTTY
	props->data.sg_flags &= ~ECHO;
#else
#ifdef  Vxw_TARGET
	props->data &= ~(OPT_ECHO);
#else
	status = -1;
#endif
#endif
#endif
	return (status);
}

static struct {
	 int rate;
	 int value; 
} speeds[] = { { 50, B50 },
		       { 75, B75 },
		       { 110, B110 },
		       { 134, B134 },
		       { 150, B150 },
		       { 200, B200 },
		       { 300, B300 },
		       { 600, B600 },
		       { 1200, B1200 },
		       { 1800, B1800 },
		       { 2400, B2400 },
		       { 4800, B4800 },
		       { 9600, B9600 },
		       { 19200, EXTA },
		       { 38400, B38400 } };

termlineSpeed (props, speed)
TermProps *props;
int speed;
{
	int status = 1;

#if defined(HAS_TERMIOS) || defined(HAS_TERMIO) || defined(HAS_SGTTY)
	int speedcode;
	int i;

#endif
#if defined(HAS_TERMIOS) || defined(HAS_TERMIO) || defined(HAS_SGTTY)
	for (i=(sizeof(speeds)/sizeof(speeds[0]))-1;i>=0;i--)
	 {
	   if (speed == speeds[i].rate)
	    { break;
	    }
	 }
	if (i >= 0)
	 { speedcode = speeds[i].value;
	 }
	else
	 { return (-1);
	 }
#endif
	if (props)
	 { 
#ifdef HAS_TERMIOS
	   cfsetospeed (&props->data, speedcode);
	   cfsetispeed (&props->data, speedcode);
#else
#ifdef HAS_TERMIO
	   props->data.c_cflag &= ~(CBAUD);
	   props->data.c_cflag |= speedcode;
#else
#ifdef HAS_SGTTY
	   props->data.sg_ispeed = speedcode;
	   props->data.sg_ospeed = speedcode;
#else
	   status = -1;
#endif
#endif
#endif
	 }
	return (status);
}

int termlineGetSpeed (props)
TermProps *props;
{
	int speedcode;
	int i;

#if defined(HAS_TERMIOS) || defined(HAS_TERMIO) || defined(HAS_SGTTY)
#ifdef HAS_TERMIOS
	speedcode = cfgetispeed (&props->data);
#else
#ifdef HAS_TERMIO
	speedcode = (props->data.c_cflag & CBAUD):

	props->data.c_cflag &= ~(CBAUD);
	props->data.c_cflag |= speedcode;
#else
#ifdef HAS_SGTTY
	speedcode = data.sg_ispeed;
#else
	speedcode = -1;
#endif
#endif
#endif
	for (i=(sizeof(speeds)/sizeof(speeds[0]))-1;i>=0;i--)
	 {
	   if (speedcode == speeds[i].value)
	    { return speeds[i].rate;
	    }
	 }
#endif
	return -1;
}
