#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>
#include <sys/poll.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/sysmacros.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "xyclkreg.h"
#include "rcisys.h"
#include "rciioctl.h"

#define STATIC

int xyclkdevflag = 0;
void *xyclki;

#define XYCLK_INTR_SPEC	0	/* device interrupt spec. */
#define XYCLK_REG_SPEC	0	/* register set to be mapped */
#define XYCLK_NULL_HANDLER (u_int(*)(caddr_t))nulldev

#define DEFAULT_RATE	5	/* default sample rate. msec. */

/* maximum number of instances supported by this driver */
#define MAX_INSTANCES		4
/* number of bits in the minor node reserved for the instance number */
#define NUM_INSTANCE_BITS	4
/* returns the instance number from a dev_t structure */
#define GET_INSTANCE(d)		(getminor(d) & ((1<<NUM_INSTANCE_BITS)-1))

/* Hardware-specific device routines */
static struct timeval xyclkHardwareRead(XyclkSoftc *);
static  int xyclkHardwareProbe (XyclkRegs *, dev_info_t *);
static void xyclkHardwareReset (XyclkRegs*, XyclkSoftc*);
static void xyclkHardwareConfig (XyclkRegs*, XyclkSoftc*);
static void xyclkHardwareIntrSetup (XyclkRegs*, XyclkSoftc*);
static void xyclkHardwareIntrEnable (XyclkRegs*, XyclkSoftc*, int);
static  int xyclkHardwareIntrHandle (XyclkRegs*, XyclkSoftc*);

/* Software support routines */
static void treqInit (XyclkSoftc*);
static int treqDispatch (XyclkSoftc*);
static int treqSchedule (void *, void (*fxn)(caddr_t), caddr_t, int);
static void treqDelete (void *, int id);

static RciTimeoutFxns timeoutFxns = {
	treqSchedule,
	treqDelete,
	0 };

static RciFxnInterface fxni;
static int fxni_set = 0;

/* Stuff relating to dynamic loading and autoconfiguration. There
   is *a lot* of boiler plate associated with this. */

static int xyclkidentify();
static int xyclkattach();
static int xyclkdetach();
static int xyclkgetinfo();
static int xyclkprobe();

STATIC int xyclkopen(dev_t *devp, int flag, int otyp, cred_t *credp);
STATIC int xyclkclose(dev_t dev, int flag, int otyp, cred_t *credp);
STATIC int xyclkread(dev_t dev, struct uio *uiop, cred_t *credp);
STATIC int xyclkwrite(dev_t dev, struct uio *uiop, cred_t *credp);
STATIC int xyclkioctl(dev_t dev, int cmd, int arg, int mode, 
			cred_t *credp, int *rvalp);
STATIC int xyclkmmap(dev_t dev, off_t off, int prot);
STATIC int xyclkchpoll(dev_t dev, short events, int anyyet,
		    short *reventsp, struct pollhead **phpp);

static u_int xyclkintr (caddr_t);
static u_int xyclksoftintr (caddr_t);

static struct cb_ops xyclk_cb_ops = {
	xyclkopen,
	xyclkclose,
	nodev,
	nodev,
	nodev,
	xyclkread,
	xyclkwrite,
	xyclkioctl,
	nodev,
	xyclkmmap,
	nodev,
	xyclkchpoll,
	nodev,
	NULL,
	D_NEW | D_MP
};

static struct dev_ops xyclk_ops = {
	DEVO_REV,
	0,
	xyclkgetinfo,
	xyclkidentify,
	xyclkprobe,
	xyclkattach,
	xyclkdetach,
	nodev,
	&xyclk_cb_ops,
	(struct bus_ops *)NULL
};

static struct modldrv modldrv = {
	&mod_driverops,
	"xyclk driver",
	&xyclk_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&modldrv,
	NULL
};

int _init (void)
{
	int status;

	ddi_soft_state_init (&xyclki, sizeof(XyclkSoftc), MAX_INSTANCES);
	status = mod_install (&modlinkage);
	if (status != 0)
	 { ddi_soft_state_fini (&xyclki);
	 }
	return (status);
}

int _fini (void)
{
	int status;

	status = mod_remove (&modlinkage);
	if (status == 0)
	 { ddi_soft_state_fini (&xyclki);
	 }
	return (status);
}

int _info (struct modinfo *modinfop)
{
	return (mod_info (&modlinkage, modinfop));
}

static int xyclkidentify (dip)
dev_info_t *dip;
{
	/* For SBus devices, the name checked by this routine must match
	   the name stored on the device itself. */

	if (strcmp (ddi_get_name(dip), "xyclk") == 0)
	 { return (DDI_IDENTIFIED);
	 }
	else
	 { return (DDI_NOT_IDENTIFIED);
	 }
}

static int xyclkgetinfo (dev_info_t *dip, ddi_info_cmd_t infocmd, 
			  void *arg, void **result)
{
	XyclkSoftc *info;
	int status;

	switch (infocmd)
	 { case DDI_INFO_DEVT2INSTANCE:
	    { *result = (void*)GET_INSTANCE((dev_t)arg);
	      status = DDI_SUCCESS;
	      break;
	    }
	   case DDI_INFO_DEVT2DEVINFO:
	    { info = ddi_get_soft_state(xyclki, GET_INSTANCE((dev_t)arg));
	      if (info == NULL)
	       { return (DDI_FAILURE);
	       }
	      *result = (void*)info->dip;
	      status = DDI_SUCCESS;
	      break;
	    }
	   default:
	    { status = DDI_FAILURE;
	      break;
	    }
	 }
	return (status);
}

static int xyclkfreeResources (dev_info_t *dip, XyclkSoftc *info)
{
	if (info->resources & XYCLK_MAIN_MINOR_CREATED)
	 { ddi_remove_minor_node (dip, NULL);
	 }
	if (info->resources & XYCLK_IHANDLER_SET)
	 { ddi_remove_intr (dip, XYCLK_INTR_SPEC, &info->iblkCookie);
	 }
	if (info->resources & XYCLK_SOFT_IHANDLER_SET)
	 { ddi_remove_softintr (info->softIntrId);
	 }
	if (info->resources & XYCLK_LOW_MUTEX_SET)
	 { mutex_destroy (&info->lowMux);
	 }
	if (info->resources & XYCLK_INTR_MUTEX_SET)
	 { mutex_destroy (&info->intrMux);
	 }
	if (info->resources & XYCLK_REGS_MAPPED)
	 { ddi_unmap_regs (dip, XYCLK_REG_SPEC, (caddr_t*)&info->regs,
		      		   /*offset=*/0, /*len=*/0);
	 }
	ddi_soft_state_free (xyclki, info->instance);
}

static int xyclkattach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	XyclkSoftc *info;
	int instance;

	static char *mainNames[] = { "0", "1", "2", "3" };
#if 0
	static char *mmapNames[] = { "mem0", "mem1", "mem2", "mem3" };
#endif
	switch (cmd)
	 { case DDI_ATTACH:
	    { instance = ddi_get_instance(dip);
	      if (instance >= MAX_INSTANCES)
	       { return (DDI_FAILURE);
	       }
	      if (ddi_soft_state_zalloc(xyclki, instance) != 0)
	       { return (DDI_FAILURE);
	       }
	      info = ddi_get_soft_state(xyclki, instance);

	      info->dip = dip;
	      info->instance = instance;
	
	      /* map device registers */
	      if (ddi_map_regs (dip, XYCLK_REG_SPEC, (caddr_t*)&info->regs,
				/*offset=*/0, /*len=*/0) != DDI_SUCCESS)
	       { xyclkfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= XYCLK_REGS_MAPPED;

	      /* configure device */
	      info->open = 0;
	      info->enabled = 1;
	      info->rate = DEFAULT_RATE;
	      info->debug = 0;
	      info->icount = 0;
	      info->hardcount = 0;
	      treqInit (info);
	      xyclkHardwareConfig (info->regs, info);
#if 0
	      RCI_setwallclock (xyclkHardwareRead);
#endif
	      /* set up interrupt-level mutex for the device. Start
		 by assigning a dummy interrupt handler so the
		 mutex is associated with the proper interrupt context,
		 without risk of an a handler getting called. */

	      if (ddi_add_intr (dip, XYCLK_INTR_SPEC,&info->iblkCookie,
		     NULL, XYCLK_NULL_HANDLER, NULL) != DDI_SUCCESS)
	       { xyclkfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }

	      mutex_init (&info->intrMux, "xyclk intr mutex", MUTEX_DRIVER,
			  (void*)info->iblkCookie);
	      ddi_remove_intr (dip, XYCLK_INTR_SPEC, info->iblkCookie);
	      info->resources |= XYCLK_INTR_MUTEX_SET;

	      /* see if the device interrupt is delivered in a high-level
		 interrupt context. If it is, we need to set up a 
		 low-level interrupt handler and mutex as well,
		 because we can't call cmn_err() or treqDispatch()
		 from a high-level interrupt context. */

	      if (ddi_intr_hilevel (dip, XYCLK_INTR_SPEC))
	       { info->hasHighLevelIntr = 1;
		 
		 /* set up low-level mutex and handler */

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			XYCLK_NULL_HANDLER, NULL) != DDI_SUCCESS)
		  { xyclkfreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 mutex_init (&info->lowMux, "xyclk low mutex", MUTEX_DRIVER,
			     (void*)info->softIblkCookie);
		 ddi_remove_softintr (info->softIntrId);
		 info->resources |= XYCLK_LOW_MUTEX_SET;

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			xyclksoftintr, (caddr_t)info) != DDI_SUCCESS)
		  { xyclkfreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 info->resources |= XYCLK_SOFT_IHANDLER_SET;

		 info->muxp = &info->lowMux;
		 info->muxh = &info->intrMux;
		 cmn_err (CE_CONT, "xyclk: interrupts are high level\n");
	       }
	      else
	       { info->hasHighLevelIntr = 0;
		 info->muxp = &info->intrMux;
		 info->muxh = 0;
		 cmn_err (CE_CONT, "xyclk: interrupts are low level\n");
	       }

	      if (ddi_add_intr (dip, XYCLK_INTR_SPEC, &info->iblkCookie,
		     &info->idevCookie, xyclkintr,(caddr_t)info) != DDI_SUCCESS)
	       { xyclkfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= XYCLK_IHANDLER_SET;

	      /* program device interrupt vector */
	      xyclkHardwareIntrSetup (info->regs, info);

	      if (ddi_create_minor_node (dip, mainNames[instance], S_IFCHR, 
		     instance, DDI_PSEUDO, 0) != DDI_SUCCESS)
	       { xyclkfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= XYCLK_MAIN_MINOR_CREATED;

#if 0
	      if (ddi_create_minor_node (dip, mmapNames[instance], S_IFCHR, 
		     instance|MMAP_MINOR, DDI_PSEUDO, 0) != DDI_SUCCESS)
	       { xyclkfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= XYCLK_MMAP_MINOR_CREATED;
#endif
	      ddi_report_dev(dip);
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }
}

static int xyclkdetach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	XyclkSoftc *info;

	switch (cmd)
	 { case DDI_DETACH:
	    { info = ddi_get_soft_state (xyclki, ddi_get_instance(dip));

	      if (fxni_set)
	       { (*fxni.setTimeoutFxns)((RciTimeoutFxns*)0);
	       }
	     /* make device quiescent */
	      xyclkHardwareIntrEnable (info->regs, info, /*enable=*/0);
	      xyclkHardwareReset (info->regs, info);
	      xyclkfreeResources (dip, info);
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }	
}

static int xyclkprobe (dev_info_t *dip)
{
        caddr_t reg_addr;
	int status = DDI_PROBE_SUCCESS;

	/* map device registers and try to contact device */
	if (ddi_map_regs (dip, XYCLK_REG_SPEC, &reg_addr, 
			       /*offset=*/0, /*len=*/0) != DDI_SUCCESS)
	 { return (DDI_PROBE_FAILURE);
	 }
	if (!xyclkHardwareProbe ((XyclkRegs*)reg_addr, dip))
	 { status = DDI_PROBE_FAILURE;
	 }
	ddi_unmap_regs (dip, XYCLK_REG_SPEC, &reg_addr,/*offset=*/0,/*len=*/0);
	return (status);
}

static u_int xyclkintr (caddr_t arg)
{
	/* Called in response to an interrupt from the XYCLK */

	XyclkSoftc *info = (XyclkSoftc*)arg;
	int needSoftIntr = 0;
	int pending;

	mutex_enter (&info->intrMux);
	pending = xyclkHardwareIntrHandle (info->regs, info);
	if (pending == 0)
	 { mutex_exit (&info->intrMux);
	   return (DDI_INTR_UNCLAIMED);
	 }

	if (pending)
	 { if (info->hasHighLevelIntr)
	    { if (!info->softIntrRunning)
	       { needSoftIntr = 1;
	       }
	      mutex_exit (&info->intrMux);
	    }
	   else
	    { mutex_exit (&info->intrMux);
	      while (treqDispatch (info))
		 ;
	    }
	 }
	else
	 { mutex_exit (&info->intrMux);
	 }
#if 0
	if (stray)
	 { /* declare a stray interrupt */
	   if (info->hasHighLevelIntr)
	    { info->stray = stray;
	      if (!info->softIntrRunning)
	       { needSoftIntr = 1;
	       }
	    }
	   else
	    { cmn_err (CE_CONT, "xyclk%d: stray interrupt(s); mask=0x%x\n",
		   		 info->instance, stray);
	    }
	 }
#endif
	if (needSoftIntr)
	 { ddi_trigger_softintr (info->softIntrId);
	 }
	return (DDI_INTR_CLAIMED);
}

static u_int xyclksoftintr (caddr_t arg)
{
	/* Used only when the device interrupts occur at high interrupt
	   priority. Calls treqDispatch() because the high-level
	   interrupt handler can't. */

	XyclkSoftc *info = (XyclkSoftc*)arg;
	int dispatched;
	int icount;

	mutex_enter (&info->intrMux);
	if (info->softIntrRunning)
	 { mutex_exit (&info->intrMux);
	   return (DDI_INTR_UNCLAIMED);	
	 }
	info->softIntrRunning = 1;
	do
	 { icount = info->icount;	/* checks if high intr occured */
	   mutex_exit (&info->intrMux);
	   dispatched = treqDispatch (info);
	   mutex_enter (&info->intrMux);
	 }
	while (dispatched || icount < info->icount);
	info->softIntrRunning = 0;
	mutex_exit (&info->intrMux);
	return (DDI_INTR_CLAIMED);	
}

int xywallclk_retry = 1;

STATIC int xyclkopen (dev_t *dev, int flag, int otyp, cred_t *crp)
{
	/* This device is an exclusive open device: only one process
	   may open it at a time. This prevents contention problems
	   and makes the driver software easier to implement. */

	XyclkSoftc *info;

	if ((info = ddi_get_soft_state(xyclki,GET_INSTANCE(*dev))) == NULL)
	 { return (ENXIO);
	 }
	if (otyp != OTYP_CHR)
	 { return (EINVAL);
	 }
	if (!info->regs)	/* paranoid; this check shouldn't be needed */
	 { return (ENXIO);
	 }
	mutex_enter (info->muxp);
	if (!info->open && info->enabled)
	 { 
	   /* set up the device, and enable interrupts */

	   if (fxni_set)
	    { timeoutFxns.arg = info;
	      (*fxni.setTimeoutFxns)(&timeoutFxns);
	    }

	   info->open = 1;
	   info->icount = 0;
	   info->hardcount = 0;
	   treqInit (info);
	   xyclkHardwareConfig (info->regs, info);
	   xyclkHardwareIntrEnable (info->regs, info, /*enable=*/1);
	 }
	mutex_exit (info->muxp);
	return (0);
}

STATIC int xyclkclose (dev_t dev, int flag, int otyp, cred_t *crp)
{
	XyclkSoftc *info;

	if ((info = ddi_get_soft_state(xyclki,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	mutex_enter (info->muxp);
	info->open = 0;
	xyclkHardwareIntrEnable (info->regs, info, /*enable=*/0);

	if (fxni_set)
	 { (*fxni.setTimeoutFxns)((RciTimeoutFxns*)0);
	 }
	mutex_exit (info->muxp);
	return (0);
}

STATIC int xyclkread (dev_t dev, uio_t *uiop, cred_t *crp)
{
	return (ENXIO);
}

STATIC int xyclkwrite (dev_t dev, uio_t *uiop, cred_t *crp)
{
	return (ENXIO);
}

STATIC int xyclkmmap (dev_t dev, off_t off, int prot)
{
	XyclkSoftc *info;

	if ((info = ddi_get_soft_state(xyclki,GET_INSTANCE(dev))) == NULL)
	 { return (-1);
	 }
	if (off < 0 || off >= sizeof(XyclkRegs))
	 { return (-1);
	 }
	return (hat_getkpfnum((char*)info->regs + off));
}

STATIC int xyclkchpoll (dev_t dev, short events, int anyyet,
		      short *reventsp, struct pollhead **phpp)
{
	return (ENXIO);
}

STATIC int xyclkioctl (dev_t dev, int cmd, int arg, int mode, 
		     cred_t *crp, int *rvalp)
{
	XyclkSoftc *info;
	int isMmapDev;

	if ((info = ddi_get_soft_state(xyclki,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }

	switch (cmd)
	 { default:
	    { return (ENOTTY);
	      break;
	    }
	   case RCIIOC_GETCLKRATE:
	    { int rate = info->rate;
	      if (copyout ((caddr_t)&rate, (caddr_t)arg, sizeof(int)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case RCIIOC_CLKSTATUS:
	    { int enabled = info->enabled;
	      if (copyout ((caddr_t)&enabled,(caddr_t)arg,sizeof(int)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case RCIIOC_SETCLKRATE:
	    { int interval = (int)arg;

	      if (interval <= 0 || (interval*500) > 0x10000)
	       { return (EINVAL);
	       }
	      mutex_enter (info->muxp);
	      info->rate = interval;
	      xyclkHardwareConfig (info->regs, info);
	      xyclkHardwareIntrEnable (info->regs, info, info->enabled);
	      mutex_exit (info->muxp);
	      break;
	    }
	   case RCIIOC_CLKENABLE:
	    {
	      mutex_enter (info->muxp);
	      if (!info->enabled)
	       { info->enabled = 1;
		 info->open = 1;
		 info->icount = 0;
		 info->hardcount = 0;
		 treqInit (info);
		 xyclkHardwareConfig (info->regs, info);
		 xyclkHardwareIntrEnable (info->regs, info, /*enable=*/1);
	       }
	      mutex_exit (info->muxp);
	      break;
	    }
	   case RCIIOC_CLKDISABLE:
	    { 
	      mutex_enter (info->muxp);
	      xyclkHardwareIntrEnable (info->regs, info, /*enable=*/0);
	      info->enabled = 0;
	      mutex_exit (info->muxp);
	      break;
	    }
	   case RCIIOC_SETCLKDEBUG:
	    { info->debug = (int)arg;
	      break;
	    }
	   case RCIIOC_SET_FXNS:
	    { 
	      if (copyin ((caddr_t)arg, (caddr_t)&fxni, sizeof(fxni)))
	       { return (EFAULT);
		 break;
	       }
	      if (fxni.setTimeoutFxns == 0)
	       { return (EFAULT);
		 break;
	       }
	      fxni_set = 1;
	      if (info->enabled)
	       { timeoutFxns.arg = info;
		 (*fxni.setTimeoutFxns)(&timeoutFxns);
	       }
	      break;
	    }
	 }
	return(0);
}

union pack {
	unsigned char c[2];
	unsigned short s;
  };
	
/* Macros for sending commands. Assumes 'regs' and 'pack' defined */
#define COMMAND_A(c)	(regs->cntrA_com = (c))
#define WRITE_A(d)	(pack.s = (d), 			\
			 regs->cntrA_data = pack.c[1], \
			 regs->cntrA_data = pack.c[0])
#define READ_A(d)	(pack.c[1] = regs->cntrA_data,	\
			 pack.c[0] = regs->cntrA_data,	\
			 (d) = pack.s)
#define COMMAND_B(c)	(regs->cntrB_com = (c))
#define WRITE_B(d)	(pack.s = (d), 			\
			 regs->cntrB_data = pack.c[1], \
			 regs->cntrB_data = pack.c[0])
#define READ_B(d)	(pack.c[1] = regs->cntrB_data,	\
			 pack.c[0] = regs->cntrB_data,	\
			 (d) = pack.s)
#define COMMAND_I(c)	(regs->uic_com = (c))
#define WRITE_I(d)	(regs->uic_data = (d))
#define READ_I(d)	((d) = regs->uic_data)

static struct timeval xyclkHardwareRead(XyclkSoftc *info)
{
	register XyclkRegs *regs;
	union pack pack;
	unsigned short count2;
	unsigned short count3;
	struct timeval tv;
	unsigned short hardcount;
	int icount;
	int rate;
	int inc;
	int usec_500;
	int extra_usec;
	int s;

	regs = info->regs;

	mutex_enter (info->muxp);

	hardcount = info->hardcount;
	rate = info->rate;
	icount = info->icount;
	if (info->muxh) mutex_enter (info->muxh);
	COMMAND_A (SAVE | 0x6);
	COMMAND_A (LOAD_PTR | HOLD_REG | CNTR2);
	READ_A (count2);
	/* If counter2 equals rate*500, it may have just wrapped
	   and we should resave counter 3 */
	if (count2 == rate*500)
	 { if (xywallclk_retry)
	    { COMMAND_A (SAVE | 0x4);
	    }
	 }
	COMMAND_A (LOAD_PTR | HOLD_REG | CNTR3);
	READ_A (count3);
	if (info->muxh) mutex_exit (info->muxh);
	mutex_exit (info->muxp);

	extra_usec = rate*500 - count2;
	if ((inc = count3 - hardcount) < 0)
	 { inc += 0x10000;
	 }
	/* total time, in units if 500 usec, is

		   2 rate icount + rate inc + extra_usec / 500

	   with remainder of extra_usec % 500. This time is
	   then converted into seconds and useconds. */

	usec_500 = 2*rate*icount + rate*inc + extra_usec/500;
	tv.tv_sec = usec_500/2000;
	tv.tv_usec = extra_usec%500 + (usec_500%2000)*500;

#if 0
	if (xywallclk_debug)
	 { printf ("count2=%d count3=%d rate=%d icount=%d hardcount=%d\n",
		    count2, count3, rate, icount, hardcount);
	   printf ("inc=%d extra_usec=%d usec_500=%d\n",
		    inc, extra_usec, usec_500);
	   printf ("tv = %d + %d\n", tv.tv_sec, tv.tv_usec);
	 }
#endif
	return (tv);	
}

static int xyclkHardwareProbe (XyclkRegs *regs, dev_info_t *dip)
{
	char c;

	if (ddi_peekc (dip, (char*)&regs->cntrA_data, &c) == DDI_FAILURE ||
	    ddi_peekc (dip, (char*)&regs->cntrA_com, &c) == DDI_FAILURE ||
	    ddi_peekc (dip, (char*)&regs->cntrB_data, &c) == DDI_FAILURE ||
	    ddi_peekc (dip, (char*)&regs->cntrB_com, &c) == DDI_FAILURE ||
	    ddi_peekc (dip, (char*)&regs->uic_data, &c) == DDI_FAILURE ||
	    ddi_peekc (dip, (char*)&regs->uic_com, &c) == DDI_FAILURE ||
	    ddi_peekc (dip, (char*)&regs->csr, &c) == DDI_FAILURE ||
	    ddi_peekc (dip, (char*)&regs->quadselA, &c) == DDI_FAILURE ||
	    ddi_peekc (dip, (char*)&regs->quadselB, &c) == DDI_FAILURE )
	 { cmn_err (CE_CONT, "xyclk: memory hole(s) at device address\n");
	   return (0);
	 }
	else
	 { return (sizeof (struct xyclkdevice));
	 }
}

static void xyclkHardwareReset(XyclkRegs *regs, XyclkSoftc *info)
{
	union pack pack;

	/* do a general rest on the device */

	if (info->muxh) mutex_enter (info->muxh);
	COMMAND_A (RESET_CNTRS);
	COMMAND_A (LOAD & 0x1f);
	COMMAND_B (RESET_CNTRS);
	COMMAND_B (LOAD & 0x1f);
	COMMAND_I (RESET_ICTRL);
	if (info->muxh) mutex_exit (info->muxh);
}

static void xyclkHardwareConfig (XyclkRegs *regs, XyclkSoftc *info)
{
	union pack pack;

	if (info->muxh) mutex_enter (info->muxh);

	/* hardware reset */
	COMMAND_A (RESET_CNTRS);
	COMMAND_A (LOAD & 0x1f);
	COMMAND_B (RESET_CNTRS);
	COMMAND_B (LOAD & 0x1f);
	COMMAND_I (RESET_ICTRL);

	/* Use counter group A. Set up the counters to run
	   at the initial default rate, and set them running */

	COMMAND_A (LOAD_PTR | MASTER_REG | CNTRL);
	WRITE_A (BCD_SCALER);

	/* Counter 1 is set to count down, divide by 4 */

	COMMAND_A (LOAD_PTR | MODE_REG | CNTR1);
	WRITE_A (SRC_F1 | REPEAT_CNT | OUT_TOGGLE);
	COMMAND_A (LOAD_PTR | LOAD_REG | CNTR1);
	WRITE_A (4);

	/* Counter 2 is set to count down, divide by rate*500
	   (interrupts are generated at half the counting rate) */

	COMMAND_A (LOAD_PTR | MODE_REG | CNTR2);
	WRITE_A (SRC_TCNM1 | REPEAT_CNT | OUT_TOGGLE);
	COMMAND_A (LOAD_PTR | LOAD_REG | CNTR2);
	WRITE_A (info->rate*500);

	/* Counter 3 counts up, and keeps track of the number of
	   interrupts times 2 */

	COMMAND_A (LOAD_PTR | MODE_REG | CNTR3);
	WRITE_A (SRC_TCNM1 | REPEAT_CNT | UP_CNT | OUT_TOGGLE);
	COMMAND_A (LOAD_PTR | LOAD_REG | CNTR3);
	WRITE_A (0);

	COMMAND_A (LOAD_ARM | 0x7);

	if (info->muxh) mutex_exit (info->muxh);
}

static void xyclkHardwareIntrSetup (XyclkRegs *regs, XyclkSoftc *info)
{
	union pack pack;

	/* program device interrupt vector */
	/* no need for interrupt mutex because interrupts can't happen yet */

	COMMAND_I (VECTOR | 0x1);
	WRITE_I (info->idevCookie.idu._idev_vector);
	COMMAND_I (MODE | HIGHEDGE_INTR);
	COMMAND_I (SELW_AC);
	WRITE_I (0xff);
	COMMAND_I (CLEAR_MR | 0x1);
	regs->csr = VME_IPL;	/* enable IPL 4 interrupts */
}

static void xyclkHardwareIntrEnable (XyclkRegs *regs, XyclkSoftc *info, 
				     int enable)
{
	/* no need for interrupt mutex because interrupts handler
	   only touches register A */
	if (enable)
	 { COMMAND_I (SELR_ARM);
	 }
	else
	 { COMMAND_I (SELR_DISARM);
	 }
}

static int xyclkHardwareIntrHandle (XyclkRegs *regs, XyclkSoftc *info)
{
	union pack pack;
	unsigned short count3;	
	int inc;
	int ticks;

	COMMAND_A (LOAD_PTR | HOLD_REG | CNTR3);
	COMMAND_A (SAVE | 0x4);
	READ_A (count3);

	if ((inc = count3 - info->hardcount) < 0)
	 { inc += 0x10000;
	 }
	ticks = (inc+1-(int)(info->hardcount%2))/2;
	info->hardcount = count3;
	info->icount += ticks;
	return (ticks);
}

static void treqInit (XyclkSoftc *info)
{
	int i;

	info->tfree = info->tbuf;
	for (i=0; i<MAX_TIMEOUTS-1; i++)
	 { info->tbuf[i].next = &info->tbuf[i+1];
	 }
	info->tbuf[MAX_TIMEOUTS-1].next = 0;
	info->treqs = 0;
}

static int treqDispatch (XyclkSoftc *info)
{
	TimeoutReq *req;

	mutex_enter (info->muxp);
	if ((req=info->treqs) && req->icount <= info->icount)
	 { info->treqs = req->next;
	 }
	else
	 { req = 0;
	 }
	mutex_exit (info->muxp);

	if (req)
	 { (*req->fxn)(req->arg);
	 }

	mutex_enter (info->muxp);
	if (req)
	 { req->next = info->tfree;
	   info->tfree = req;
	 }
	mutex_exit (info->muxp);

	return (req ? 1 : 0);
}

static int treqSchedule (void *handle, void (*fxn)(caddr_t), 
			  caddr_t arg, int ticks)
{
	XyclkSoftc *info = (XyclkSoftc*)handle;
	TimeoutReq *req;
	TimeoutReq *prev;
	TimeoutReq *node;

	mutex_enter (info->muxp);
	if ((req = info->tfree) == NULL)
	 { /* no free requests */
	   mutex_exit (info->muxp);
	   return (0);
	 }
	else
	 { req->icount = info->icount + ticks;
	   req->arg = arg;
	   req->fxn = fxn;
	   prev = 0;
	   for (node=info->treqs; node && req->icount >= node->icount;
		node=node->next)
	    { prev = node;
	    }
	   req->next = node;
	   if (prev)
	    { prev->next = req;
	    }
	   else
	    { info->treqs = req;
	    }
	 }
	mutex_exit (info->muxp);
	return (req-info->tbuf+1);
}

static void treqDelete (void *handle, int id)
{
	XyclkSoftc *info = (XyclkSoftc*)handle;
	TimeoutReq *req;
	TimeoutReq *prev;
	TimeoutReq *node;

	req = &info->tbuf[id-1];
	mutex_enter (info->muxp);
	prev = 0;
	for (node=info->treqs; node; node=node->next)
	 { if (node == req)
	    { if (prev)
	       { prev->next = node->next;
	       }
	      else
	       { info->treqs = node->next;
	       }
	      node->next = info->tfree;
	      info->tfree = node;
	      mutex_exit (info->muxp);
	      return;
	    }
	   prev = node;
	 }
	mutex_exit (info->muxp);
}
